unit Tomas;

(***********************************************************************)
(*  This software is free software;                                    *)
(*  You may distribute or change the software as you wish, but         *)
(*  this header must always be present.                                *)
(*  No one is allowed to claim copyright on this source code.          *)
(*  All precompiled software using this unit should be accompanied     *)
(*  by this source code.                                               *)
(*                                                                     *)
(*  Original contents on http://www.tomasweb.com                       *)
(*  Contact: Hans P.M. Veeke         H.P.M.Veeke@wbmt.tudelft.nl       *)
(***********************************************************************)

(***********************************************************************)
(* Update history                                                      *)
(* 24-11-2001: Queuelengths changed to LongWords                       *)
(* 26-08-2001: StandBy mechanism corrected.                            *)
(*             Random generator corrected (could end with zero value)  *)
(*             Changes for speed implemented                           *)
(*             SetMessageProcessInterval added                         *)
(*             TomasForm not anymore 'stayontop'                       *)
(*             Trace messages extended with clocktime stamp            *)
(* 18-05-2001: MoveTo corrected to reach endpoint exactly              *)
(* 14-05-2001: Cancel corrected.                                       *)
(* 18-04-2001: Animation provisional implemented                       *)
(* 17-04-2001: Distributed facilities extended                         *)
(* 07-04-2001: Standby bug corrected                                   *)
(* 29-10-2000: Mean function added to distributions.                   *)
(*             Tabledistribution corrected.                            *)
(*   Thanks to Mark Duinkerken.                                        *)
(* 26-09-2000: TomasServer methods implemented                         *)
(* 24-09-2000: Step mode flag before startup corrected                 *)
(* 11-09-2000: Pause method added as a synonym of interrupt            *)
(*             Copy, clear and move methods added to TomasQueue        *)
(*             Status and eventtime added to elements-form             *)
(* 15-08-2000: interrupt method correctly implemented                  *)
(*             ResetTomas added.                                       *)
(* 20-07-2000: facilities for concurrent development implemented       *)
(*             and tested.                                             *)
(* 20-06-2000: incorrect error message when resuming a claiming        *)
(*             element solved                                          *)
(* 19-06-2000: to reach complete analogy with TomasProcess:            *)
(*             StopSimulation(T) added                                 *)
(*             SimulationRunStatus added                               *)
(*             Global variable TheGraphicsForm for correct closing     *)
(* 06-06-2000: StandBy corrected                                       *)
(*             StopAndLeave added (analogy of FinishAndDestroy)        *)
(*             FinishAndDestroy corrected                              *)
(* 31-05-2000: Normal distribution corrected                           *)
(* 29-05-2000: next changes made because of TomasResources             *)
(*             - Changed Claimroom to public queue                     *)
(*             - Attribute Claims of TomasElement is made public       *)
(*                                                                     *)
(* Original version: 28-05-2000                                        *)
(***********************************************************************)


interface

uses
  TomasProcess,
  TomasAnimation,
  GMP,
  Elements,
  Queues,
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  ExtCtrls,
  Menus,
  Buttons,
  Gauges,
  ComCtrls;

type

  TomasQueue = class;
  TomasElement = class;
  TomasShape = class;
  TNode = class;
  TQueueLink = class;
  TTomasQuestion = class;
  TCameraElement = class;
  SortProcedure = Function(C1,C2: Pointer): Boolean;
  ReceiveProcedure = Procedure(Msg: String);
  TimerProcedure = Procedure;
  DistributionType = (Uniform,Exponential,Normal,Table);
  DistributionTableType = (Cumulative,Discrete);
  SimulationRunStatus = (RunInit,RunFinished,RunInterrupted,Running,RunStepping);
  TomasRenderMode = (rmWireframe, rmSolid, rmSolidShade,
   rmSolidTexture, rmShadedTexture);

  TTomasForm = class(TForm)
    TracePanel: TPanel;
    TraceMemo: TMemo;
    TraceCheckBox: TCheckBox;
    TomasMainMenu: TMainMenu;
    StartItem: TMenuItem;
    InterruptItem: TMenuItem;
    ResumeItem: TMenuItem;
    QuitItem: TMenuItem;
    StepModeCheckBox: TCheckBox;
    StepButton: TSpeedButton;
    StepBevel: TBevel;
    SettingsPanel: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    ServerAddressCombo: TComboBox;
    BreakEdit: TEdit;
    OutputCombo: TComboBox;
    SettingOKButton: TBitBtn;
    SettingsCancelButton: TBitBtn;
    TomasStatusBar: TStatusBar;
    HeaderControl1: THeaderControl;
    Splitter1: TSplitter;
    AnimationCheckBox: TCheckBox;
    MessageCheckBox: TCheckBox;
    ClockTimeBox: TCheckBox;
    StandByCheckBox: TCheckBox;
    procedure GoCreateTomasForm(Sender: TObject);
    procedure TraceCheckClick(Sender: TObject);
    procedure StepModeCheckClick(Sender: TObject);
    procedure ExecuteNextStep(Sender: TObject);
    procedure GoHideSettings(Sender: TObject);
    procedure GoAcceptSettings(Sender: TObject);
    procedure CheckFormResize(Sender: TObject);
    procedure GoStartSimulation(Sender: TObject);
    procedure GoResumeSimulation(Sender: TObject);
    procedure GoQuitSimulation(Sender: TObject);
    procedure CheckClosing(Sender: TObject; var Action: TCloseAction);
    procedure GoInterruptSimulation(Sender: TObject);
    procedure CheckSpaceKey(Sender: TObject; var Key: Char);
    procedure CheckStatusBarMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure GoResizeTraceMemo(Sender: TObject);
    procedure GoEditTheSettings(HeaderControl: THeaderControl;
      Section: THeaderSection);
    procedure GoSetMessageCheck(Sender: TObject);
    procedure GoShowTomasForm(Sender: TObject);
    procedure TraceStandByCheckClick(Sender: TObject);
  protected
    FTrace: Boolean;
    FTraceStandBy: Boolean;
    FStepMode: Boolean;
  private
    CloseWindow: Boolean;
    Procedure SetTracememo;
  public
    Procedure SetTrace(OnOff: Boolean);
    Procedure SetTraceStandBy(OnOff: Boolean);
    Procedure SetStepMode(OnOff: Boolean);
    Function GetServerAddress: String;
    Procedure SetServerAddress(SrvAddress: String);
    Procedure GoSetAnimationOnOrOff;
    Property Trace: Boolean read FTrace write SetTrace;
    Property TraceStandBy: Boolean read FTraceStandBy write SetTraceStandBy;
    Property StepMode: Boolean read FStepMode write SetStepMode;
  end;

  {TomasElement is the basic class of Tomas.
  Active Elements should override the process-method.
  Remember to put this method in the 'published' section
  of your Element-definition. }
TomasElement = Class(TomasProcessElement)
  Private
    FQueues: TNode;
    FDuration: Double;
    FInClaimRoom: Boolean;
  Published
    Constructor Create(S: String);
    Destructor Destroy; override;
    {Processing methods}
    Procedure Start(T: Double);
    Procedure Stop;
    Procedure StopAndLeave;
    Procedure Resume(T: Double);
    Procedure Cancel;
    Procedure Pause;
    Procedure Interrupt;
    {Queueing methods}
    Procedure EnterQueue(Q: TomasQueue);
    Procedure EnterQueueInFront(Q: TomasQueue);
    Procedure EnterQueueSorted(Q: TomasQueue; SortingProc: SortProcedure);
    Procedure LeaveQueue(Q: TomasQueue);
    Function Successor(Q: TomasQueue): Pointer;
    Function Predecessor(Q: TomasQueue): Pointer;
    Function IsInQueue(Q: TomasQueue): Boolean;
    {Information methods}
    Function QueueTime(Q: TomasQueue): Double;
    Function StatusToString: String;
    Procedure WriteTrace(S: String);
  End;

TomasShape = Class(TomasElement)
  Private
    TC_Moving: Boolean;
    TC_StartMove: Double;
    TC_PosX: Single;
    TC_PosY: Single;
    TC_PosZ: Single;
    TC_AngleX: Single;
    TC_AngleY: Single;
    TC_AngleZ: Single;
    TC_ToX: Single;
    TC_ToY: Single;
    TC_ToZ: Single;
    TC_Appearance: TCAppearance;
    TC_Camera: TCameraElement;
  Published
    Constructor Create(S: String);
    Destructor Destroy; override;
    Procedure GetPosition(Var X,Y,Z: Single);
    Function GetPositionX: Single;
    Function GetPositionY: Single;
    Function GetPositionZ: Single;
    Procedure GetOrientation(Var AX,AY,AZ: Single);
    Procedure AddBoxObject(X,Y,Z: Single; rm: TomasRenderMode;
     Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
    Procedure AddCircleObject(R: Single; Rm: TomasRenderMode;
     Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
    Procedure ReadShapeFromFile(FileName: String);
    Procedure Show(X,Y,Z: Single);
    Procedure Turn(Ax,Ay,Az: Single);
    Procedure Hide;
    Procedure MoveTo(X,Y,Z,Step,Speed: Single);
    Procedure DMoveTo(X,Y,Z,Step,Speed: Single);
    Procedure Bend2D(AX,AY,AZ,Radius,AStep,Speed: Single);
    Procedure ConnectWith(Shape: TomasShape; X,Y,Z: Single);
    Procedure DisConnect;
    Procedure SetRenderMode(rm: TomasRenderMode);
    Procedure SetColor(Color: TColor);
  End;


TCameraElement = class(TomasElement)
  Private
  TC_Position: Array[1..3] Of Single;
  TC_Rotation: Array[1..3] Of Single;
  TC_Dx,TC_Dy,TC_Dz: Single;
  TC_Connection: TomasShape;
  Published
  Procedure Process; Override;
  Procedure ConnectTo(Shape: TomasShape; Dx,Dy,Dz,Rx,Ry,Rz: Single);
  Procedure Disconnect;
  Procedure SetRotation(X,Y,Z: Single);
  Procedure SetPosition(X,Y,Z: Single);
  End;

{TomasQueue is the basic class to contain subsets of
 Elements. Normally a TomasQueue will be used to represent
 waiting-rows, stock or work-in-progress.
 Only TomasElements are allowed to be entered in a TomasQueue.
 Automatic gathering of statistical data is provided
 about queuecontents and waiting-times.}
TomasQueue = Class
  Private
    FItems: TNode;
    FQLink: TQueueLink;
    FName: String;
    FLength: LongWord;
    FResetTime: Double;
    FLastChange: Double;
    FMeanLength: Double;
    FMinLength: LongWord;
    FMaxLength: LongWord;
    FMeanWT: Double;
    FMinWT: Double;
    FMaxWT: Double;
    FDirectPassed: LongWord;
    FTotalPassed: LongWord;
  Published
    Constructor Create(Name: String);
    Destructor Destroy; Override;
    Procedure AddToTail(Comp: TomasElement); virtual;
    Procedure AddToHead(Comp: TomasElement); virtual;
    Procedure AddSorted(Comp: TomasElement; SortingProc: SortProcedure); virtual;
    Procedure AddSortedOn(Comp: TomasElement; SortingPar: Double); virtual;
    Procedure AddBefore(Comp,BeforeComp: TomasElement);
    Procedure AddAfter(Comp,AfterComp: TomasElement);
    Procedure Remove(Comp: TomasElement); virtual;
    Function FirstElement: Pointer;
    Function LastElement: Pointer;
    Function Successor(Comp: TomasElement): Pointer;
    Function Predecessor(Comp: TomasElement): Pointer;
    Function Contains(Comp: TomasElement): Boolean;
    Function IndexOf(TElm: TomasElement): Integer;
    Function Element(Ix: Integer): Pointer;
    function ElementWithName(Nm: String): Pointer;
    Procedure ResetStatistics;
    Function Copy: TomasQueue;
    Procedure Clear;
    Function Move: TomasQueue;
    property Name: String read FName;
    property Length: LongWord read FLength;
    property ResetTime: Double read FResetTime;
    property MeanLength: Double read FMeanLength;
    property MinLength: LongWord read FMinLength;
    property MaxLength: LongWord read FMaxLength;
    property MeanWT: Double read FMeanWT;
    property MinWT: Double read FMinWT;
    property MaxWT: Double read FMaxWT;
    property DirectPassed: LongWord read FDirectPassed;
    property TotalPassed: LongWord read FTotalPassed;
  Private
    Procedure CalculateMeanLength;
  end;


TDistribution = class
  Private
    Seed: Integer;
    DMean: Double;
  Published
    Constructor Create(DSeed: Integer);
    Function Sample: Double; virtual;
    Function Mean: Double;
    Function CurrentSeed: Integer;
  End;

TExponentialDistribution = class(TDistribution)
  Published
    Constructor Create(Seed: Integer; Mn: Double);
    Function Sample: Double; override;
  End;

TUniformDistribution = class(TDistribution)
  Private
    LBound: Double;
    UBound: Double;
  Published
    Constructor Create(Seed: Integer; Lb,Ub: Double);
    Function Sample: Double; override;
  End;

TNormalDistribution = class(TDistribution)
  Private
    Sigma: Double;
  Published
    Constructor Create(Seed: Integer; Mn,S: Double);
    Function Sample: Double; override;
    End;

TTableDistribution = class(TDistribution)
  Private
    TableType: DistributionTableType;
    XFValues: TList;
  Published
    Constructor Create(Seed: Integer; DType: DistributionTableType);
    Procedure AddValue(X,F: Double); {F <= 1; If cumulative last F <= 1}
    Function Sample: Double;  override;
  End;

TInputDistribution = class
  Private
    InputType: DistributionType;
    Distribution: TDistribution;
  published
    Constructor Create(DString: String);
    Function Sample: Double;
    Function Mean: Double;
    End;

TTomasQuestion = class
  ID: Integer;
  Phrase: String;
  Source: String;
  End;

TMsgString = class
  Private
  MsgS: String;
  SubMsgS: String;
  Procedure SetNextStringPart;
  Public
  MsgError:Boolean;
  published
  Constructor Create(Msg: String);
  Function GetBoolean: Boolean;   {TRUE if y,Y,t,T,1}
  Function GetSmallInt: SmallInt; {-32768,..,32767}
  Function GetLongInt: LongInt;   {-2147483648,..,2147483647}
  Function GetByte: Byte;         {0,..,255}
  Function GetWord: Word;         {0,..,65535}
  Function GetInteger: Integer;   {-2147483648,..,2147483647}
  Function GetCardinal: Cardinal; {0,..,2147483647}
  Function GetReal: Real;         {6 bytes}
  Function GetDouble: Double;     {8 bytes}
  Function GetSingle: Single;     {4 bytes}
  Function GetString: String;
  End;

  TNode = class
    Item: TomasElement;
    Queue: TomasQueue;
    ArrivalTime: Double;
    SortingPar: Double;
    QNext: TNode;
    QPred: TNode;
    ENext: TNode;
    EPred: TNode;
    Destructor Destroy; override;
    Procedure ElementToQueueBefore(Tail: TNode);
    Procedure ElementToQueueAfter(Head: TNode);
    Procedure QueueToElementBefore(Tail: TNode);
    End;

  TQueueLink = class
    Ref: TomasQueue;
    Next: TQueueLink;
    Pred: TQueueLink;
    Destructor Destroy; override;
    End;

  AnswerProcedure = Procedure(Q: TMsgString; Source: String);


Procedure StartSimulation;
Procedure StopSimulation(T: Double);
Procedure InterruptSimulation;
Procedure ResumeSimulation;
Procedure SetSimulationSpeed(Speed: Integer);
Function TNow: Double;
Function CurrentElement: TomasElement;
Function RunStatus: SimulationrunStatus;
Procedure Hold(T: Double);
Procedure StandBy;
Procedure Suspend;
Procedure Finish;
Procedure FinishAndDestroy;
Procedure ResetTomas;
Function Rand(Var L: Integer): Double;
procedure SetTimeDisplayInterval(T: Double);
Procedure SetMessageProcessInterval(MPInt: Integer);
Function Data: ElementStatus;
Function Claiming: ElementStatus;
Function Passive: ElementStatus;
Function Interrupted: ElementStatus;
Function TimeScheduled: ElementStatus;
Function StateScheduled: ElementStatus;
Function Current: ElementStatus;
Procedure WriteToTrace(S: String);
Procedure SetModelName(MName: String);
Procedure SetServerName(Svname: String);
procedure SendMessageToModel(Model,Msg: String);
Procedure SendMessageToServer(ServerMsg: String);
procedure SetReceive(RcvProc: ReceiveProcedure);
Procedure SetTimerProc(TimProc: TimerProcedure);
Procedure SetAnswerProc(AnswerProc: AnswerProcedure);
Procedure SetOutputDestination(S: String);
Procedure ConnectServer;
Procedure DisConnectServer;
Procedure StartLogging;
Procedure StopLogging;
Procedure SendQuestionTo(Model: String; Var Question: String);
Procedure SendReply(Destination,Reply: String);
Procedure SetAnimationDisplayInterval(T: Double);
Procedure MoveCamera(CKey: Word; CShift: TShiftstate; Delta: Single);
Procedure SetScreenString(S: String);
Procedure SetAnimation;
Procedure SetBackGround(P: TBackGround);
Procedure SetBackGroundColor(Color: TColor);
Procedure DrawLine(X1,Y1,Z1,X2,Y2,Z2: Single; Color: TColor);
procedure GoSaveQueues(Qf: String);



var
  TomasForm: TTomasForm;
  TheGraphicsForm: TForm;
  ClaimRoom: TomasQueue;
  TomasCamera: TCameraElement;

implementation

{$R *.DFM}

Type

  TElementClass = class
    Num: Integer;
    Name: String;
    Method: Pointer;
    Next: TElementClass;
    End;

  TXFItem = class
    X_Value: Double;
    F_Value: Double;
    End;

Var
  ElementClassHead: TElementClass;
  ElementClassTail: TElementClass;
  NrOfQueues: Integer;
  DoTrace: Boolean;
  ElementRoom: TomasQueue;
  QueueRoom: TQueueLink;
  TimeDisplayInterval: Double;
  LastEditTime: Double;
  OutputFilename: String;
  OutputDestinationFile: Boolean;
  OutputFile: TextFile;
  PReceiveMessage: ReceiveProcedure;
  PTimer: TimerProcedure;
  PAnswerQuestion: AnswerProcedure;
  QuestionID: Integer;
  AnimationOn: Boolean;
  AnimationDisplayInterval: Double;
  ScreenString: String;
  LastRefreshTime: Double;

(****************************************************************)
(*            general supporting routines                       *)
(****************************************************************)

Function TFloatToStrF(V: Extended; L,NrD: Integer): String;
  Var
    S: String;
  Begin
  S:=FloatToStrF(V,ffFixed,L,NrD);
  Result:=Copy('                    ',1,L - Length(S)) + S;
  End;

Procedure SetAnimationDisplayInterval(T: Double);
  Begin
  AnimationDisplayInterval:=T;
  End;

Procedure WriteToTrace(S: String);
  Begin
  If OutputDestinationFile Then
    begin
    Append(OutputFile);
    WriteLn(OutputFile,S);
    CloseFile(OutputFile);
    End
  Else
    Begin
    If TomasForm.TraceMemo.Lines.Count>=500 Then
      Begin
      TomasForm.TraceMemo.Lines.Delete(0);
      End;
    TomasForm.TraceMemo.Lines.Append(S);
    End;
  End;

Procedure CheckTrace(TStr,CStr,AStr: String; T: Double);
  Var
    S: String;
  Begin
  If (TomasForm.Trace Or TomasForm.TraceStandBy) And DoTrace Then
    Begin
    If OutputDestinationFile Then {File}
      Begin
      Append(OutputFile);
      If T = -1 Then
        Begin
        WriteLn(OutputFile,TimeToStr(Time) + '/' + TStr + ' ' + CStr + ' ' + AStr);
        End
      Else
        Begin
        Str(T:10:2,S);
        WriteLn(OutputFile,TimeToStr(Time) + '/' + TStr + ' ' + CStr + ' ' + AStr + ' ' + S);
        End;
      CloseFile(OutputFile);
      End
    Else
      Begin
      If TomasForm.TraceMemo.Lines.Count>=500 Then
        Begin
        TomasForm.TraceMemo.Lines.Delete(0);
        End;
      If T = -1 Then
        Begin
        TomasForm.TraceMemo.Lines.Append(TStr + ' ' + CStr + ' ' + AStr);
        End
      Else
        Begin
        Str(T:10:2,S);
        TomasForm.TraceMemo.Lines.Append(TStr + ' ' + CStr + ' ' + AStr + ' ' + S);
        End;
      End;
    End;
  End;    {CheckTrace}

Procedure GoShowElements;
  Var
    Element: TomasElement;
    EQLink: TNode;
  Begin
  TomasForm.GoInterruptSimulation(Nil);
  If NOT(Assigned(ElementForm)) Then
    Begin
    Application.CreateForm(TElementForm,ElementForm);
    End
  Else
    Begin
    ElementForm.ClearElements;
    End;
  Element:=ElementRoom.FirstElement;
  While Element <> Nil Do
    Begin
    ElementForm.AddElement(Element.Name,Element.ArrivalTime,
     Element.StatusToString,Element.EventTime);
    EQLink:=Element.FQueues.ENext.ENext;   {skip Elementroom}
    While EQLink <> Element.FQueues Do
      Begin
      ElementForm.AddQueue(EQLink.Queue.Name,EQLink.ArrivalTime);
      EQLink:=EQLink.ENext;
      End;
    Element:=ElementRoom.Successor(Element);
    End;
  ElementForm.WindowState:=wsNormal;
  ElementForm.Show;
  ElementForm.BringToFront;
  ElementForm.SetFocus;
  End;

Procedure GoShowQueues;
  Var
    QQLink: TQueueLink;
  Begin
  TomasForm.GoInterruptSimulation(Nil);
  If NOT(Assigned(QueueForm)) Then
    Begin
    Application.CreateForm(TQueueForm,QueueForm);
    End
  Else
    Begin
    QueueForm.ClearQueues;
    End;
  QQLink:=QueueRoom.Next;
  While QQLink <> QueueRoom Do
    Begin
    With QQLink.Ref Do
      Begin
      QueueForm.AddQueue(Name,Length,MeanLength,MaxLength,TotalPassed,
       DirectPassed,MeanWT,MaxWT);
      End;
    QQLink:=QQLink.Next;
    End;
  QueueForm.WindowState:=wsNormal;
  QueueForm.Show;
  QueueForm.BringToFront;
  QueueForm.SetFocus;
  End;

Function LString(IStr: String; L: Integer): String;
  Var
    S: String;
    i: Integer;
  Begin
  S:='';
  For i:=1 To L - Length(IStr) Do
    Begin
    S:=S + ' ';
    End;
  S:=S + IStr;
  Result:=S;
  End;

Function RString(IStr: String; R: Integer): String;
  Var
    S: String;
    i: Integer;
  Begin
  S:=IStr;
  For i:=1 To R - Length(IStr) Do
    Begin
    S:=S + ' ';
    End;
  Result:=S;
  End;
  

procedure GoSaveQueues(Qf: String);
  Var
    QQLink: TQueueLink;
    QFile: textFile;
  Begin
  TomasForm.GoInterruptSimulation(Nil);
  AssignFile(QFIle,Qf);
  ReWrite(QFile);
  WriteLn(QFile,'Name        Length    Mean      Max       Passed    Direct    '
   + 'MeanWT    MaxWT');
  QQLink:=QueueRoom.Next;
  While QQLink <> QueueRoom Do
    Begin
    With QQLink.Ref Do
      Begin
      WriteLn(QFile,RString(Name,12),LString(IntToStr(Length),10),
       LString(FormatFloat('0.00',MeanLength),10),
       LString(IntToSTr(MaxLength),10),
       LString(IntToStr(TotalPassed),10),
       LString(IntToSTr(Directpassed),10),
       LString(FormatFloat('0.00',MeanWT),10),
       LString(FormatFloat('0.00',MaxWT),10));
      End;
    QQLink:=QQLink.Next;
    End;
  CloseFile(QFile);
  End;

(****************************************************************)
(*            public routines                                   *)
(****************************************************************)

Procedure StartSimulation;
  Begin
  SetServer(TomasForm.GetServerAddress);
  StartProcessSimulation;
  TomasForm.Close;
  If TomasForm.CloseWindow Then
    Begin
    TomasForm.Close;
    End;
  End;

Procedure StopSimulation(T: Double);
  Begin
  SetStopTime(T);
  End;

Procedure InterruptSimulation;
  Begin
  TomasForm.GoInterruptSImulation(Nil);
  End;

Procedure ResumeSimulation;
  Begin
  TomasForm.GoResumeSimulation(Nil);
  End;

Procedure SetSimulationSpeed(Speed: Integer);
  Begin
  GoSetSimulationSpeed(Speed);
  End;

Function TNow: Double;
  Begin
  Result:=GetNow;
  End;

Function CurrentElement: TomasElement;
  Begin
  Result:=TomasElement(CurrentProcessElement);
  End;

Function RunStatus: SimulationRunStatus;
  Begin
  Result:=SimulationRunStatus(GetSimulationStatus);
  End;

Function TNowString: String;
  Var
    S: String;
  Begin
  Str(TNow:12:2,S);
  Result:=S;
  End;

{With MoveCamera the user is able to program
 the camerabehaviour. Input is the keycombination used
 when the user moves the camera interactively.
 The following combinations are allowed:
   VK_UP: move into Z-direction forward;
   VK_DOWN: move into Z-direction backward;
   VK_LEFT: move into X-direction left;
   VK_RIGHT: move into X-direction right;
   Ctrl + VK_UP: move into Y-direction Up;
   Ctrl + VK_DOWN: move into Y-direction Down;
   Alt + VK_UP: Rotate around X-axis clockwise;
   Alt + VK_Down: Rotate around X-axis counterclockwise;
   Alt + VK_Right: Rotate around Y-axis counterclockwise;
   Alt + VK_Left: Rotate around Y-axis clockwise;
 The movement is done in units of Delta. By adding the
 shift-state the movement is done in units of 10 * Delta}
Procedure MoveCamera(CKey: Word; CShift: TShiftstate; Delta: Single);
  Var
    DeltaPos: Single;
  Begin
  If TomasCamera.TC_Connection <> Nil Then
    begin
    Exit;
    End;
  While TomasAnimationForm.KeyIsDown Do
    Begin
    With TomasCamera,TomasAnimationForm Do
      Begin
      GetCameraPosition(TC_Position[1],TC_Position[2],TC_Position[3]);
      If ssShift in CShift Then
        begin
        DeltaPos:=Delta * 10;
        End
      Else
        Begin
        DeltaPos:=Delta;
        End;
      If NOT(ssCtrl in CShift) And NOT(ssAlt in CShift) Then
        begin
        Case CKey Of
          VK_UP: SetCameraPosition(TC_Position[1],TC_Position[2],TC_Position[3]-DeltaPos);
          VK_DOWN: SetCameraPosition(TC_Position[1],TC_Position[2],TC_Position[3]+DeltaPos);
          VK_LEFT: SetCameraPosition(TC_Position[1]-DeltaPos,TC_Position[2],TC_Position[3]);
          VK_RIGHT:SetCameraPosition(TC_Position[1]+DeltaPos,TC_Position[2],TC_Position[3]);
          End;
        End
      Else
        Begin
        If ssCtrl in CShift Then
          Begin
          case CKey Of
            VK_UP: SetCameraPosition(TC_Position[1],TC_Position[2]+DeltaPos,TC_Position[3]);
            VK_DOWN:SetCameraPosition(TC_Position[1],TC_Position[2]-DeltaPos,TC_Position[3]);
            End;
          End
        Else
          Begin
          If ssAlt in CShift Then
            Begin
            GetCameraRotation(TC_Rotation[1],TC_Rotation[2],TC_Rotation[3]);
            Case CKey Of
              VK_UP:SetCameraRotation(TC_Rotation[1] - Deltapos,TC_Rotation[2],TC_Rotation[3]);
              VK_Down:SetCameraRotation(TC_Rotation[1] + DeltaPos,TC_Rotation[2],TC_Rotation[3]);
              VK_Right:SetCameraRotation(TC_Rotation[1],TC_Rotation[2]-DeltaPos,TC_Rotation[3]);
              VK_Left:SetCameraRotation(TC_Rotation[1],TC_Rotation[2]+DeltaPos,TC_Rotation[3]);
              End;
            End;
          End;
        End;
      Refresh(Nil);
      End;
    Application.ProcessMessages;
    End;
  End;   {MoveCamera}

{SetScreenString sets the string displayed in the
 upperleft corner of the AnimationDisplay.
 Tomas automatically adds the simulationtime in
 front of it}
Procedure SetScreenString(S: String);
  Begin
  ScreenString:=S;
  End;  {SetScreenString}

{SetAnimation creates and shows the animation-
 form of Tomas}
Procedure SetAnimation;
  Begin
  AnimationOn:=TRUE;
  TomasForm.GoSetAnimationOnOrOff;
  End;  {SetAnimation}

{SetBackGround sets the procedure to draw
 a background-image on the animationscreen}
Procedure SetBackGround(P: TBackGround);
  Begin
  TomasAnimationForm.PBackGround:=P;
  End;  {SetBackGround}

{SetBackGroundColor sets the backgroundcolor
 of the animationscreen}
Procedure SetBackGroundColor(Color: TColor);
  Begin
  TomasAnimationForm.SetBackGroundColor(Color);
  End;   {SetBackGroundColor}

{DrawLine draws a line from (X1,Y1,Z1) to (X2,Y2,Z2)
 with Color 'Color'}
Procedure DrawLine(X1,Y1,Z1,X2,Y2,Z2: Single; Color: TColor);
  Begin
  TomasAnimationForm.DrawLine(X1,Y1,Z1,X2,Y2,Z2,Color);
  End;   {DrawLine}

Procedure Hold(T: Double);
  Begin
  If CurrentElement = Nil Then
    Begin
    MessageDlg('Error(Hold): no current Element',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If T < 0 Then
    Begin
    MessageDlg('Error(Hold): negative time' + FloatToStrF (T, ffFixed, 5, 3),
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,CurrentElement.Name,'holds until',TNow + T);
  HoldProcess(T);
  End;   {Hold}

Procedure StandBy;
  Begin
  If CurrentElement = Nil Then
    Begin
    MessageDlg('Error(StandBy): no current Element',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If (TomasForm.TraceStandBy) And DoTrace Then
    CheckTrace(TNowString,CurrentElement.Name,'goes into StandBy',-1);
  StandByProcess;
  End;  {StandBy}

Procedure Suspend;
  Begin
  If CurrentElement = Nil Then
    Begin
    MessageDlg('Error(Passivate): no current Element',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If CurrentElement.FInClaimRoom Then
    Begin
    CurrentElement.Status:=CLAIMING;
    If TomasForm.Trace And DoTrace Then
      CheckTrace(TNowString,CurrentElement.Name,'starts claiming',-1);
    End
  Else
    Begin
    CurrentElement.Status:=PASSIVE;
    If TomasForm.Trace And DoTrace Then
      CheckTrace(TNowString,CurrentElement.Name,'suspends',-1);
    End;
  SuspendProcess;
  End;   {Suspend}


Procedure Finish;
  Begin
  If CurrentElement = Nil Then
    Begin
    MessageDlg('Error(finish): no current Element',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,CurrentElement.Name,' finished',-1);
  FinishProcess;
  End;   {Finish}

Procedure FinishAndDestroy;
  Begin
  If CurrentElement = Nil Then
    Begin
    MessageDlg('Error(FinishAndDestroy): no current Element',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,CurrentElement.Name,' finished and destroyed',-1);
  FinishProcessAndDestroy;
  End;  {FinishAndDestroy}

Function Data: ElementStatus;
  Begin
  Result:=TC_DATA;
  End; {Data}

Function Claiming: ElementStatus;
  Begin
  Result:=TC_CLAIMING;
  End;  {Claiming}

Function Passive: ElementStatus;
  begin
  Result:=TC_PASSIVE;
  End;  {Passive}

Function Interrupted: ElementStatus;
  Begin
  Result:=TC_INTERRUPTED;
  End;  {Interrupted}

Function TimeScheduled: ElementStatus;
  begin
  Result:=TC_TIMESCHEDULED;
  End;  {TimeScheduled}

Function StateScheduled: ElementStatus;
  Begin
  Result:=TC_STATESCHEDULED;
  End;  {StateScheduled}

Function Current: ElementStatus;
  Begin
  Result:=TC_CURRENT;
  End;  {StateScheduled}

procedure SetTimeDisplayInterval(T: Double);
  Begin
  TimeDisplayInterval:=T;
  End;

Procedure SetMessageProcessInterval(MPInt: Integer);
  Begin
  SetProcessingInterval(MPInt);
  End;

Procedure SetModelName(MName: String);
  Begin
  If MName = '' Then
    Begin
    MName := LowerCase(ExtractFileName(Application.ExeName));
    MName := Copy(MName,1,Length(MName) - 4);
    End;
  GoSetModelName(MName);
  TomasForm.Caption:=ProcessVersion + ' Object oriented Modeling And Simulation: ' + MName;
  End;

Procedure SetServerName(SVname: String);
  Begin
  If GetSimulationStatus = S_INIT Then
    Begin
    SetServer(SvName);
    TomasForm.HeaderControl1.Sections[0].Text:='ServerAddress: ' + SvName;
    End;
  End;

procedure SendMessageToModel(Model,Msg: String);
  Begin
  If GetServer <> '' Then
    begin
    SendModelMessage(Model,Msg);
    End;
  End;

Procedure SendMessageToServer(ServerMsg: String);
  Begin
  {Messages can be: 'finish', 'start: x', 'stop', 'time' and 'clients'}
  If GetServer <> '' Then
    Begin
    SendModelMessage('$$server',ServerMsg);
    End;
  End;

procedure SetReceive(RcvProc: ReceiveProcedure);
  Begin
  PReceiveMessage:=RcvProc;
  End;

Procedure SetTimerProc(TimProc: TimerProcedure);
  begin
  PTimer:=TimProc;
  End;  

Procedure SetAnswerProc(AnswerProc: AnswerProcedure);
  Begin
  PAnswerQuestion:=AnswerProc;
  End;

Procedure ConnectServer;
  Begin
  CheckServer;
  End;

Procedure DisconnectServer;
  Begin
  DisConnectFromServer;
  End;

Procedure StartLogging;
  begin
  DoStartLogging;
  End;

Procedure StopLogging;
  Begin
  DoStopLogging;
  End;

Procedure SendQuestionTo(Model: String; Var Question: String);
  Begin
  DoSendQuestionTo(Model,Question);
  End;

Procedure SendReply(Destination,Reply: String);
  Begin
  DoSendReplyTo(Destination,Reply,QuestionID);
  End;

(****************************************************************)
(*            Methods of TomasElement                           *)
(****************************************************************)

Constructor TomasElement.Create(S: String);
  Var
    CurClass: TElementClass;
    MAddress: Pointer;
  Begin
  S:=Trim(S);
  MAddress:=MethodAddress('Process');
  CurClass:=ElementClassHead;
  While (CurClass<>Nil) And ((CurClass.Method <> MAddress)
   Or (CurClass.Name <> S)) Do
    Begin
    CurClass:=CurClass.Next;
    End;
  If CurClass = Nil Then
    Begin
    CurClass:=TElementClass.Create;
    CurClass.Num:=0;
    CurClass.Name:=Trim(S);
    CurClass.Method:=MAddress;
    If ElementClassTail = Nil Then
      Begin
      ElementClassHead:=CurClass;
      ElementClassTail:=CurClass;
      CurClass.Next:=Nil;
      End
    Else
      begin
      ElementClassTail.Next:=CurClass;
      ElementClassTail:=CurClass;
      CurClass.Next:=Nil;
      End;
    End;
  Inc(CurClass.Num);
  If (CurClass.Num = 1) Then
    Begin
    inherited Create(S,TNow);
    End
  Else
    Begin
    inherited Create(S + Trim(IntToStr(CurClass.Num)),TNow);
    End;
  FQueues:=TNode.Create;
  FQueues.Item:=Self;
  FQueues.Queue:=Nil;
  FQueues.QNext:=Nil;
  FQueues.QPred:=Nil;
  FQueues.ENext:=FQueues;
  FQueues.EPred:=FQueues;
  FDuration:=-1;
  DoTrace:=FALSE;
  EnterQueue(ElementRoom);
  DoTrace:=TRUE;
  FInClaimRoom:=FALSE;
  if TomasForm.Visible Then
   TomasForm.TomasStatusBar.Panels[1].Text:='Elements: ' + IntToStr(ElementRoom.Length);
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Name,'created',-1);
  End;   {TomasElement.Create}

Destructor TomasElement.Destroy;
  Var
    CNode: TNode;
  begin
  CNode:=FQueues.ENext;
  While CNode<>FQueues Do
    Begin
    LeaveQueue(CNode.Queue);
    CNode:=FQueues.ENext;
    End;
  FQueues.Free;
  if TomasForm.Visible Then
   TomasForm.TomasStatusBar.Panels[1].Text:='Elements: ' + IntToStr(ElementRoom.Length);
  inherited Destroy;
  End;      {TomasElement.Destroy}

Procedure TomasElement.Start(T: Double);
  Begin
  If Status > PASSIVE Then
    Begin
    MessageDlg(TNowString + ' Error(StartProcess): ' + Name + ' already scheduled',
     mtError,[mbOK],0);
    RunError(255);
    End;
  If Status > Data Then
    Begin
    MessageDlg(TNowString + ' Error(StartProcess): ' + Name + ' passive or waiting',
     mtError,[mbOK],0);
    RunError(255);
    End;
  If TomasForm.Trace And DoTrace Then
      CheckTrace(TNowString,Name,'starts at',T);
  ScheduleProcess(T);
  End;   {TomasElement.Start}

Procedure TomasElement.Resume(T: Double);
  Begin
  If (Status = DATA) Or (Status > INTERRUPTED) Then
    Begin
    MessageDlg(TNowString + ' Error(ResumeProcess): ' + Name + ' not passive',
     mtError,[mbOK],0);
    RunError(255);
    End;
  If T < TNow Then
    Begin
    MessageDlg(TNowString + ' Error(ResumeProcess): ' + Name + ' resumes before now',
     mtError,[mbOK],0);
    RunError(255);
    End;
  If (Status = Interrupted) Then
    Begin
    ScheduleProcess(T + FDuration);
    If TomasForm.Trace And DoTrace Then
      CheckTrace(TNowString,Name,'resumes',-1);
    FDuration:=-1;
    End
  Else
    Begin
    ScheduleProcess(T);
    If TomasForm.Trace And DoTrace Then
      CheckTrace(TNowString,Name,'resumes at ',T);
    End;
  End;     {TomasElement.Resume}

Procedure TomasElement.Stop;
  Begin
  FDuration:=-1;
  If CurrentElement = Self Then
    Begin
    Finish;
    End
  Else
    Begin
    If Scheduled Then
      Begin
      {add interpolation in case of TomasShape!!!!!}
      If TomasForm.Trace And DoTrace Then
        CheckTrace(TNowString,Name,' finished',-1);
      RemoveProcess(TRUE);
      End
    Else
      Begin
      If (Status = DATA) Then
        Begin
        MessageDlg('Error(Stop): State of ' + Name + ' is DATA',
         mtError,[mbOK],0);
        Runerror(255);
        End;
      If (Status = Passive) Then
        Begin
        CancelResume;
        End;
      End;
    End;
  End;  {TomasElement.Stop}

Procedure TomasElement.StopAndLeave;
  Begin
  If CurrentElement = Self Then
    Begin
    FinishAndDestroy;
    End
  Else
    Begin
    If Scheduled Then
      Begin
      {add interpolation in case of TomasShapeElement!!!!!}
      If TomasForm.Trace And DoTrace Then
        CheckTrace(TNowString,Name,' finished',-1);
      RemoveProcess(TRUE);
      End
    Else
      Begin
      If (Status = DATA) Then
        Begin
        MessageDlg('Error(Stop): State of ' + Name + ' is DATA',
         mtError,[mbOK],0);
        Runerror(255);
        End;
      End;
    Free;
    End;
  End;  {TomasElement.StopAndLeave}

Procedure TomasElement.Cancel;
  Begin
  If (Status < PASSIVE) Or (Status = CURRENT) Then
    Begin
    MessageDlg('Error(Cancel): ' + Name + ' is not scheduled or '
     + 'interrupted',mtError,[mbOK],0);
    RunError(255);
    End
  Else
    Begin
    If Scheduled Then
      Begin
      CancelProcess;
      End;
    Status:=Data;
    FDuration:=-1;
    End;
  End;   {TomasElement.Cancel}

Procedure TomasElement.Pause;
  Begin
  interrupt;
  End;

Procedure TomasElement.Interrupt;
  Begin
  If (Status <= INTERRUPTED) Then
    Begin
    MessageDlg('Error(Interrupt): ' + Name + ' is not scheduled '
     + 'or already interrupted',mtError,[mbOK],0);
    RunError(255);
    End
  Else
    Begin
    If Scheduled Then
      Begin
      If Status = TimeScheduled Then
        Begin
        FDuration:=EventTime - TNow;
        End
      Else
        Begin
        FDuration:=-2;
        End;
      InterruptProcess;
      If TomasForm.Trace And DoTrace Then
        CheckTrace(TNowString,Name,' interrupted',-1);
      End;
    End;
  End;   {TomasElement.Interrupt}


Procedure TomasElement.EnterQueue(Q: TomasQueue);
  Begin
  Q.AddToTail(Self);
  End;   {TomasElement.EnterQueue}

Procedure TomasElement.EnterQueueInFront(Q: TomasQueue);
  Begin
  Q.AddToHead(Self);
  End;  {TomasElement.EnterQueueInFront}

Procedure TomasElement.EnterQueueSorted(Q: TomasQueue; SortingProc: SortProcedure);
  Begin
  Q.AddSorted(Self,SortingProc);
  End;   {TomasElement.EnterQueueSorted}

Procedure TomasElement.LeaveQueue(Q: TomasQueue);
  Begin
  Q.Remove(Self)
  End;   {TomasElement.LeaveQueue}

Function TomasElement.Successor(Q: TomasQueue): Pointer;
  Begin
  Result:=Q.Successor(Self);
  End;   {TomasElement.Successor}

Function TomasElement.Predecessor(Q: TomasQueue): Pointer;
  Begin
  Result:=Q.Predecessor(Self);
  End;   {TomasElement.Predecessor}

Function TomasElement.IsInQueue(Q: TomasQueue): Boolean;
  Begin
  Result:=Q.Contains(Self);
  End;   {TomasElement.IsInQueue}


Function TomasElement.QueueTime(Q: TomasQueue): Double;
  Var
    CNode: TNode;
  Begin
  CNode:=FQueues.ENext.ENext;
  While (CNode <> FQueues) And (CNode.Queue <> Q) Do
    Begin
    CNode:=CNode.ENext;
    End;
  If CNode = FQueues Then
    Begin
    MessageDlg('Error(GetQueueTime): ' + Name + ' is not in queue ' + Q.Name,
     mtError,[mbOK],0);
    Runerror(255);
    Result:=-1;
    End
  Else
    Begin
    Result:=CNode.ArrivalTime;
    End;
  End;  {TomasElement.QueueTime}

Function TomasElement.StatusToString: String;
  Begin
  Case Status Of
    TC_DATA: Result:='Data';
    TC_CLAIMING: Result:='Claiming';
    TC_PASSIVE: Result:='Passive';
    TC_INTERRUPTED: Result:='Interrupted';
    TC_TIMESCHEDULED: Result:='Scheduled on Time';
    TC_STATESCHEDULED: Result:='Scheduled on State';
    TC_CURRENT: Result:='Current';
    End;
  End;   {TomasElement.StatusToString}


Procedure TomasElement.WriteTrace(S: String);
  Begin
  If OutputDestinationFile Then
    begin
    Append(OutputFile);
    WriteLn(OutputFile,TNowString + ' ' + Name + ' ' + S);
    CloseFile(OutputFile);
    End
  Else
    Begin
    With TomasForm.TraceMemo Do
      begin
      If Lines.Count>=500 Then
        Begin
        Lines.BeginUpdate;
        Lines.Delete(0);
        Lines.EndUpDate;
        End;
      Lines.Append(TNowString + ' '
       + Name + ' ' + S);
      End;
    End;
  End;

(****************************************************************)
(*            Methods of TomasShape                             *)
(****************************************************************)


Constructor TomasShape.Create(S: String);
  Begin
  Inherited Create(S);
  TC_Appearance:=TCAppearance.Create;
  TC_Moving:=FALSE;
  TC_Camera:=Nil;
  TC_StartMove:=0;
  TC_ToX:=0;
  TC_ToY:=0;
  TC_ToZ:=0;
  TC_PosX:=0;
  TC_PosY:=0;
  TC_PosZ:=0;
  TC_AngleX:=0;
  TC_AngleY:=0;
  TC_AngleZ:=0;
  End; {TomasShape.Create}

Destructor TomasShape.Destroy;
  Begin
  TC_Appearance.Destroy;
  inherited Destroy;
  End; {TomasShape.Destroy}

{GetPosition returns the current position
 of the Shape (if necessary by means of interpolation}
Procedure TomasShape.GetPosition(Var X,Y,Z: Single);
  begin
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  If TC_Moving Then
    Begin
    X:=TC_PosX + (TC_ToX - TC_PosX)*(TNow - TC_StartMove)/(EventTime - TC_StartMove);
    Y:=TC_PosY + (TC_ToY - TC_PosY)*(TNow - TC_StartMove)/(EventTime - TC_StartMove);
    Z:=TC_PosZ + (TC_ToZ - TC_PosZ)*(TNow - TC_StartMove)/(EventTime - TC_StartMove);
    End
  Else
    Begin
    X:=TC_PosX;
    Y:=TC_PosY;
    Z:=TC_PosZ;
    End;
  End; {TomasShape.GetPosition}

{GetPositionX returns the current X-position
 of the Shape (if necessary by means of interpolation}
Function TomasShape.GetPositionX: Single;
  Begin
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  If TC_Moving Then
    Begin
    Result:=TC_PosX + (TC_ToX - TC_PosX)*(TNow - TC_StartMove)/(EventTime - TC_StartMove);
    End
  Else
    Begin
    Result:=TC_PosX;
    End;
  End;  {TomasShape.GetPositionX}

{GetPositionY returns the current Y-position
 of the Shape (if necessary by means of interpolation}
Function TomasShape.GetPositionY: Single;
  Begin
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  If TC_Moving Then
    Begin
    Result:=TC_PosY + (TC_ToY - TC_PosY)*(TNow - TC_StartMove)/(EventTime - TC_StartMove);
    End
  Else
    Begin
    Result:=TC_PosY;
    End;
  End;  {TomasShape.GetPositionY}

{GetPositionZ returns the current Z-position
 of the Shape (if necessary by means of interpolation}
Function TomasShape.GetPositionZ: Single;
  Begin
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  If TC_Moving Then
    Begin
    Result:=TC_PosZ + (TC_ToZ - TC_PosZ)*(TNow - TC_StartMove)/(EventTime - TC_StartMove);
    End
  Else
    Begin
    Result:=TC_PosZ;
    End;
  End;  {TomasShape.GetPositionZ}

{GetOrientation returns the current orientation
 (in degrees) of the TomasShape.}
Procedure TomasShape.GetOrientation(Var AX,AY,AZ: Single);
  Begin
  AX:=TC_Appearance.GetAngleX;
  AY:=TC_Appearance.GetAngleY;
  AZ:=TC_Appearance.GetAngleZ;
  End;  {TomasShape.GetOrientation}

{AddBoxObject adds a box-object to the Shape
 The center of the box is at position X,Y,Z relative to the
 origin of the Shape. The rendermode is given by rm,
 the Color is defined by Color.
 Rx,Ry and Rz represent the axis around which the box should
 be turned 'Angle' degrees. Dx, Dy, Dz represent the
 displacement relative to the box's origin. The origin of
 the box is its center}
Procedure TomasShape.AddBoxObject(X,Y,Z: Single; rm: TomasRenderMode;
 Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Begin
  TC_Appearance.AddBoxObject(X,Y,Z,TRenderMode(rm),Color,Rx,Ry,Rz,Angle,Dx,Dy,Dz);
  End; {TomasShape.AddBoxObject}

{AddCircleObject adds a circle to the Shape with radius R.
 The rendermode is given by rm, the Color is defined by Color.
 Rx,Ry and Rz represent the axis around which the box should
 be turned 'Angle' degrees. Dx, Dy, Dz represents the
 displacement relative to the Shape's origin}
Procedure TomasShape.AddCircleObject(R: Single; Rm: TomasRenderMode;
 Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Begin
  TC_Appearance.AddCircleObject(R,TRenderMode(rm),Color,Rx,Ry,Rz,Angle,Dx,Dy,Dz);
  End; {TomasShape.AddCircleObject}

{ReadShapeFromFile reads a complete appearance from
 file FileName. The format of data is not yet formalised.
 Just for example some '.geo' files are added to the
 Tomas-examples}
Procedure TomasShape.ReadShapeFromFile(FileName: String);
  Begin
  TC_Appearance.LoadFromFile(FileName);
  End; {TomasShape.ReadShapeFromFile}

{The Show-method can be used to show a Shape.
 The Shape will be shown with its origin
 at world-position (X,Y,Z) in its current orientation}
Procedure TomasShape.Show(X,Y,Z: Single);
  Begin
  TC_PosX:=X;
  TC_PosY:=Y;
  TC_PosZ:=Z;
  TC_Appearance.Show(X,Y,Z,0);
  If TC_Camera <> Nil Then
    Begin
    With TC_Camera Do
      Begin
      TC_Position[1]:=X + TC_Dx;
      TC_Position[2]:=Y + TC_Dy;
      TC_Position[3]:=Z + TC_Dz;
      TomasAnimationForm.SetCameraPosition(TC_Position[1],TC_Position[2],TC_Position[3]);
      End;
    End;
  End; {TomasShape.Show}

{The Turn-method turns the Shape. Ax, Ay or Az
 denote the number of degrees to turn around each axis.
 A positive value denotes a counterclockwise turn.}
Procedure TomasShape.Turn(Ax,Ay,Az: Single);
  Begin
  If Ax <> 0 Then
    Begin
    TC_Appearance.Show(1,0,0,Ax);
    End;
  If Ay <> 0 Then
    Begin
    TC_Appearance.Show(0,1,0,Ay);
    End;
  If Az <> 0 Then
    Begin
    TC_Appearance.Show(0,0,1,Az);
    End;
  TC_AngleX:=TC_Appearance.GetAngleX;
  TC_AngleY:=TC_Appearance.GetAngleY;
  TC_AngleZ:=TC_Appearance.GetAngleZ;
  if TC_Camera <> Nil Then
    begin
    With TC_Camera Do
      Begin
      TomasAnimationForm.GetCameraRotation(TC_Rotation[1],TC_Rotation[2],TC_Rotation[3]);
      TomasAnimationForm.SetCameraRotation(TC_Rotation[1] + Ax,
       TC_Rotation[2] + Ay, TC_Rotation[3] + Az);
      End;
    End;
  End; {TomasShape.Turn}

{the Hide method hides the Shape}
Procedure TomasShape.Hide;
  begin
  TC_Appearance.Hide;
  End;  {TomasShape.Hide}

{The MoveTo-method moves the Shape from the current
 position in a straight line to (X,Y,Z) with speed 'Speed'
 and shows it every 'Step'-distance units}
Procedure TomasShape.MoveTo(X,Y,Z,Step,Speed: Single);
  Var
    XDist: Single;
    YDist: Single;
    ZDist: Single;
    Dist: Single;
    DeltaX: Single;
    DeltaY: Single;
    DeltaZ: Single;
    EndPointReached: Boolean;
  Begin
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  XDist:=X - TC_PosX;
  YDist:=Y - TC_PosY;
  ZDist:=Z - TC_PosZ;
  Dist:=Sqrt(XDist*XDist + YDist * YDist + ZDist * ZDist);
  TC_Moving:=TRUE;
  If (Dist > 0) And (Speed > 0) And (Step > 0) Then
    Begin
    If Step > Dist Then
      Begin
      Step:=Dist;
      End;
    DeltaX:=(Step/Dist) * XDist;
    DeltaY:=(Step/Dist) * YDist;
    DeltaZ:=(Step/Dist) * ZDist;
    EndPointReached:=FALSE;
    While NOT(EndPointReached) Do
      Begin
      TC_StartMove:=TNow;   {for interpolation purposes}
      TC_ToX:=TC_PosX + DeltaX;
      TC_ToY:=TC_PosY + DeltaY;
      TC_ToZ:=TC_PosZ + DeltaZ;
      Hold(Step/Speed);
      TC_PosX:=TC_ToX;
      TC_PosY:=TC_ToY;
      TC_PosZ:=TC_ToZ;
      Show(TC_PosX,TC_PosY,TC_PosZ);
      EndPointReached := (TC_PosX = X) And (TC_PosY = Y)
       And (TC_PosZ = Z) Or (Dist = 0);
      If NOT(EndPointReached) Then
        begin
        If Dist >= Step Then
          Begin
          Dist:=Dist - Step;
          End
        Else
          Begin
          DeltaX:=X - TC_PosX;
          DeltaY:=Y - TC_PosY;
          DeltaZ:=Z - TC_PosZ;
          Dist:=0;
          End;
        End;
      End;
    End;
  TC_Moving:=FALSE;
  End; {TomasShape.MoveTo}

{The DMoveTo-method moves the TomasShape (X,Y,Z)-
 distance units from the current position with speed 'Speed'
 and shows it every 'Step'- distance units}
Procedure TomasShape.DMoveTo(X,Y,Z,Step,Speed: Single);
  Begin
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  MoveTo(TC_PosX + X,TC_PosY + Y, TC_PosZ + Z, Step, Speed);
  End;  {TomasShape.DMoveTo}

{The Bend2D-method performs a circular Bend with the
 Shape. AX, AY or AZ denote the number of degrees to
 turn. Because of the 2-dimensional nature only one of
 these can be nonzero. If the value is positive the Bend
 will be counterclockwise, if negative it will be clockwise.
 'Radius' is the Bendradius, while the bend is performed with
 speed 'Speed'. Every 'AStep' distance-units the shape
 will be refreshed on the screen.}
Procedure TomasShape.Bend2D(AX,AY,AZ,Radius,AStep,Speed: Single);
  Var
    Xm: Single;
    Ym: Single;
    Zm: Single;
    ADest: Single;
    Delta: Integer;
    TotAngle: Single;
    StepDist: Single;
  Begin
  Delta:=0;
  ADest:=0;
  Xm:=0;
  Ym:=0;
  Zm:=0;
  TC_Appearance.GetPosition(TC_PosX,TC_PosY,TC_PosZ);
  TC_AngleX:=TC_Appearance.GetAngleX;
  TC_AngleY:=TC_Appearance.GetAngleY;
  TC_AngleZ:=TC_Appearance.GetAngleZ;
  TC_Moving:=TRUE;
  TotAngle:=0;
  If Radius=0 Then
    begin
    StepDist:=Abs(AStep)/Speed;
    End
  Else
    Begin
    StepDist:=Pi * Radius * AStep / 180.0;
    End;
  If AX<>0 Then
    Begin
    Delta:=Ord(Ax>0) - Ord(Ax<0);
    Xm:=TC_PosX;
    Ym:=TC_PosY + Delta * Radius * Sin(TC_AngleX * Pi / 180.0);
    Zm:=TC_PosZ - Delta * Radius * Cos(TC_AngleX * Pi / 180.0);
    ADest:=Ax;
    End;
  If AY<>0 Then
    Begin
    Delta:=Ord(Ay>0) - Ord(Ay<0);
    Xm:=TC_PosX - Delta * Radius * Cos(TC_AngleY * Pi / 180.0);
    Ym:=TC_PosY;
    Zm:=TC_PosZ + Delta * Radius * Sin(TC_AngleY * Pi / 180.0);
    ADest:=AY;
    End;
  If AZ<>0 Then
    Begin
    Delta:=Ord(AZ>0) - Ord(AZ<0);
    Xm:=TC_PosX - Delta * Radius * Cos(TC_AngleZ * Pi / 180.0);
    Ym:=TC_PosY + Delta * Radius * Sin(TC_AngleZ * Pi / 180.0);
    Zm:=TC_PosZ;
    ADest:=AZ;
    End;
  While TotAngle <> Abs(ADest) Do
    Begin
    TC_StartMove:=TNow;
    TC_AngleX:=TC_AngleX + Ord(Ax<>0) * Delta * Astep;
    TC_AngleY:=TC_AngleY + Ord(Ay<>0) * Delta * Astep;
    TC_AngleZ:=TC_AngleZ + Ord(Az<>0) * Delta * Astep;
    TotAngle:=TotAngle + AStep;
    If Abs(TotAngle) > Abs(ADest) Then
      Begin
      TC_AngleX := TC_AngleX - Ord(Ax<>0) * (TotAngle - ADest);
      TC_AngleY := TC_AngleY - Ord(Ay<>0) * (TotAngle - ADest);
      TC_AngleZ := TC_AngleZ - Ord(Az<>0) * (TotAngle - ADest);
      TotAngle := ADest;
      End;
    If AX<>0 Then
      Begin
      TC_ToX:=Xm;
      TC_ToY:=Ym - Delta * Radius * Sin(TC_AngleX * Pi / 180.0);
      TC_ToZ:=Zm + Delta * Radius * Cos(TC_AngleX * Pi / 180.0);
      End;
    If AY<>0 Then
      Begin
      TC_ToX:=Xm + Delta * Radius * Cos(TC_AngleY * Pi / 180.0);
      TC_ToY:=Ym;
      TC_ToZ:=Zm - Delta * Radius * Sin(TC_AngleY * Pi / 180.0);
      End;
    If AZ<>0 Then
      Begin
      TC_ToX:=Xm + Delta * Radius * Cos(TC_AngleZ * Pi / 180.0);
      TC_ToY:=Ym - Delta * Radius * Sin(TC_AngleZ * Pi / 180.0);
      TC_ToZ:=Zm;
      End;
    Hold(StepDist/Speed);
    TC_PosX:=TC_ToX;
    TC_PosY:=TC_ToY;
    TC_PosZ:=TC_ToZ;
    If AX<>0 Then
      Begin
      Turn(-Delta * AStep,0,0);
      End;
    If AY<>0 Then
      Begin
      Turn(0,Delta * AStep,0);
      End;
    If AZ<>0 Then
      Begin
      Turn(0,0,-Delta * AStep);
      End;
    Show(TC_PosX,TC_PosY,TC_PosZ);
    End;
  TC_Moving:=FALSE;
  End; {TomasShape.Bend2D}

{The ConnectWith-method combines this Shape
 with the Shape 'ShapeElm'. This Shape
 is placed at offset X,Y,Z of the center of the other
 Shape. Every movement of Shape will now automatically
 be performed on this Shape}
Procedure TomasShape.ConnectWith(Shape: TomasShape; X,Y,Z: Single);
  Begin
  Shape.TC_Appearance.Add(TC_Appearance,X,Y,Z);
  End; {TomasShape.ConnectWith}

{The DisConnect-method releases this Shape of the
 Shape it has been connected with}
Procedure TomasShape.DisConnect;
  Begin
  TC_Appearance.UnTie;
  End; {TomasShape.DisConnect}

{SetRenderMode sets the rendering mode of the Shape.}
Procedure TomasShape.SetRenderMode(rm: TomasRenderMode);
  Begin
  TC_Appearance.SetRenderMode(TrenderMode(rm));
  End; {TomasShape.SetRenderMode}

{SetColor sets the color of the Shape}
Procedure TomasShape.SetColor(Color: TColor);
  Begin
  TC_Appearance.SetColor(Color);
  End; {TomasShapeElement.SetColor}

(****************************************************************)
(*            Methods of TCameraElement                         *)
(****************************************************************)

Procedure TCameraElement.Process;
  Begin
  While TomasAnimationForm.KeyIsDown Do
    Begin
    Application.ProcessMessages;
    End;
  End;   {TCameraElement.Process}

Procedure TCameraElement.ConnectTo(Shape: TomasShape; Dx,Dy,Dz,Rx,Ry,Rz: Single);
  Begin
  TC_Connection:=Shape;
  Shape.TC_Camera:=Self;
  Shape.GetPosition(TC_Position[1],TC_Position[2],TC_Position[3]);
  TC_Rotation[1]:=Shape.TC_Appearance.GetAngleX;
  TC_Rotation[2]:=Shape.TC_Appearance.GetAngleY;
  TC_Rotation[3]:=Shape.TC_Appearance.GetAngleZ;
  TomasAnimationForm.SetCameraPosition(TC_Position[1] + Dx, TC_Position[2] + Dy, TC_Position[3] + Dz);
  TC_Dx:=Dx;
  TC_Dy:=Dy;
  TC_Dz:=Dz;
  TomasAnimationForm.SetCameraRotation(TC_Rotation[1] + Rx,TC_Rotation[2] + Ry,TC_Rotation[3] + Rz);
  End;  {TCameraElement.ConnectTo}

Procedure TCameraElement.Disconnect;
  Begin
  TC_Connection.TC_Camera:=Nil;
  TC_Connection:=Nil;
  End;  {TCameraElement.DisConnect}

Procedure TCameraElement.SetRotation(X,Y,Z: Single);
  Begin
  TomasAnimationForm.SetCameraRotation(X,Y,Z);
  End;   {TCameraElement.SetRotation}

Procedure TCameraElement.SetPosition(X,Y,Z: Single);
  Begin
  TomasAnimationForm.SetCameraPosition(X,Y,Z);
  End;  {TCameraElement.SetPosition}

(****************************************************************)
(*            Methods of TomasQueue                             *)
(****************************************************************)

Constructor TomasQueue.Create(Name: String);
  Begin
  Inherited Create;
  FName:=Name;
  FLength:=0;
  FMinWT:=NEVER;
  FMaxWT:=0;
  FItems:=TNode.Create;
  FItems.Queue:=Self;
  FItems.Item:=Nil;
  FItems.ArrivalTime:=-1;
  FItems.QNext:=FItems;
  FItems.QPred:=FItems;
  FItems.ENext:=Nil;
  FItems.EPred:=Nil;
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Name,'created',-1);
  If DoTrace Then
    Begin
    Inc(NrOfQueues);
    if TomasForm.Visible Then
     TomasForm.TomasStatusBar.PAnels[2].Text:='Queues: ' + IntToStr(NrOfQueues);
    End;
  FQLink:=TQueueLink.Create;
  FQLink.Ref:=Self;
  FQLink.Next:=QueueRoom;
  FQLink.Pred:=QueueRoom.Pred;
  FQLink.Pred.Next:=FQLink;
  QueueRoom.Pred:=FQLink;
  FLastChange:=TNow;
  ResetStatistics;
  End;   {TomasQueue.Create}

Destructor TomasQueue.Destroy;
  Var
    NextItem: TNode;
  Begin
  {remove items}
  NextItem:=FItems.QNext;
  While NextItem<>FItems Do
    Begin
    NextItem.Free;
    NextItem:=FItems.QNext;
    End;
  NextItem.Free;
  If Assigned(FQLink) Then
    Begin
    FQLink.Free;
    End;
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Name,'destroyed',-1);
  Dec(NrOfQueues);
  if TomasForm.Visible Then
   TomasForm.TomasStatusBar.Panels[2].Text:='Queues: ' + IntToStr(NrOfQueues);
  inherited Destroy;
  End;   {TomasQueue.Destroy}


Procedure TomasQueue.AddToTail(Comp: TomasElement);
  Var
    NewItem: TNode;
  Begin
  CalculateMeanLength;
  NewItem:=TNode.Create;
  NewItem.Queue:=Self;
  NewItem.Item:=Comp;
  NewItem.ArrivalTime:=TNow;
  NewItem.SortingPar:=0;
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=TRUE;
  Inc(FLength);
  NewItem.ElementToQueuebefore(FItems);  {connect to tail of Queuechain}
  NewItem.QueueToElementBefore(COmp.FQueues);    {connect to tail of ElementQchain}
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,'to tail of '+ Name,-1);
  If FLength > FMaxLength Then
    Begin
    FMaxLength:=FLength;
    End;
  End;   {TomasQueue.AddToTail}

Procedure TomasQueue.AddToHead(Comp: TomasElement);
  Var
    NewItem: TNode;
  Begin
  CalculateMeanLength;
  NewItem:=TNode.Create;
  NewItem.Queue:=Self;
  NewItem.Item:=Comp;
  NewItem.ArrivalTime:=TNow;
  NewItem.SortingPar:=0;
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=TRUE;
  Inc(FLength);
  NewItem.ElementToQueueAfter(FItems);   {connect to front of Queuechain}
  NewItem.QueueToElementBefore(Comp.FQueues);    {connect to tail of ElementQchain}
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,'to head of '+ Name,-1);
  If FLength > FMaxLength Then
    Begin
    FMaxLength:=FLength;
    End;
  End;   {TomasQueue.AddToHead}

Procedure TomasQueue.AddSorted(Comp: TomasElement; SortingProc: SortProcedure);
  Var
    NewItem: TNode;
    NextItem: TNode;
  Begin
  CalculateMeanLength;
  NewItem:=TNode.Create;
  NewItem.Queue:=Self;
  NewItem.Item:=Comp;
  NewItem.ArrivalTime:=TNow;
  NewItem.SortingPar:=0;
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=TRUE;
  Inc(FLength);
  {connect to front of Queuechain}
  NextItem:=FItems.QNext;
  While (NextItem<>FItems) And
   NOT(SortingProc(NewItem.Item,NextItem.Item)) Do
    Begin
    NextItem:=NextItem.QNext;
    End;
  NewItem.ElementToQueueBefore(NextItem);
  NewItem.QueueToElementBefore(COmp.FQueues);
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,'to sort of '+ Name,-1);
  If FLength > FMaxLength Then
    Begin
    FMaxLength:=FLength;
    End;
  End;   {TomasQueue.AddSorted}

Procedure TomasQueue.AddSortedOn(Comp: TomasElement; SortingPar: Double);
  Var
    NewItem: TNode;
    NextItem: TNode;
  Begin
  CalculateMeanLength;
  NewItem:=TNode.Create;
  NewItem.Queue:=Self;
  NewItem.Item:=Comp;
  NewItem.ArrivalTime:=TNow;
  NewItem.SortingPar:=SortingPar;
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=TRUE;
  Inc(FLength);
  {connect to front of Queuechain}
  NextItem:=FItems.QNext;
  While (NextItem<>FItems) And
   (NewItem.SortingPar >= NextItem.SortingPar) Do
    Begin
    NextItem:=NextItem.QNext;
    End;
  NewItem.ElementToQueueBefore(NextItem);
  NewItem.QueueToElementBefore(COmp.FQueues);
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,'to sort of '+ Name,-1);
  If FLength > FMaxLength Then
    Begin
    FMaxLength:=FLength;
    End;
  End;   {TomasQueue.AddSortedOn}

Procedure TomasQueue.AddBefore(Comp,BeforeComp: TomasElement);
  Var
    NewItem: TNode;
    ContItem: TNode;
  Begin
  CalculateMeanLength;
  NewItem:=TNode.Create;
  NewItem.Queue:=Self;
  NewItem.Item:=Comp;
  NewItem.ArrivalTime:=TNow;
  NewItem.SortingPar:=0;
  Inc(FLength);
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=TRUE;
  {connect before beforecomp in queuechain}
  ContItem:=BeforeComp.FQueues.ENext;
  While (ContItem<>BeforeComp.FQueues) And (ContItem.Queue<>Self) Do
    Begin
    ContItem:=ContItem.ENext;
    End;
  If ContItem = BeforeComp.FQueues Then
    Begin
    MessageDlg(TNowString + ' Error(AddBefore): BeforeComp '
     + BeforeComp.Name + ' not in ' + Name,mtError,[mbOK],0);
    RunError(255);
    End;
  NewItem.ElementToQueueBefore(ContItem);
  NewItem.QueueToElementBefore(Comp.FQueues);
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,' put before ' + BeforeComp.Name +
   ' in '+ Name,-1);
  If FLength > FMaxLength Then
    Begin
    FMaxLength:=FLength;
    End;
  End;  {TomasQueue.AddBefore}

Procedure TomasQueue.AddAfter(Comp,AfterComp: TomasElement);
  Var
    NewItem: TNode;
    ContItem: TNode;
  Begin
  CalculateMeanLength;
  NewItem:=TNode.Create;
  NewItem.Queue:=Self;
  NewItem.Item:=Comp;
  NewItem.ArrivalTime:=TNow;
  NewItem.SortingPar:=0;
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=TRUE;
  Inc(FLength);
  ContItem:=AfterComp.FQueues.ENext;
  While (ContItem<>AfterComp.FQueues) And (ContItem.Queue<>Self) Do
    Begin
    ContItem:=ContItem.ENext;
    End;
  If ContItem = AfterComp.FQueues Then
    Begin
    MessageDlg(TNowString + ' Error(AddAfter): AfterComp '
     + AfterComp.Name + ' not in ' + Name,mtError,[mbOK],0);
    RunError(255);
    End;
  NewItem.ElementToQueueAfter(ContItem);
  {connect to tail of ElementQchain}
  NewItem.QueueToElementBefore(Comp.FQueues);
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,' put after ' + AfterComp.Name +
   ' in '+ Name,-1);
  If FLength > FMaxLength Then
    Begin
    FMaxLength:=FLength;
    End;
  End;  {TomasQueue.AddAfter}

Function TomasQueue.FirstElement: Pointer;
  Var
    FirstItem: TNode;
  Begin
  FirstItem:=FItems.QNext;
  If FirstItem = FItems Then
    Begin
    Result:=Nil;
    End
  Else
    Begin
    Result:=FirstItem.Item;
    End;
  End;   {TomasQueue.FirstElement}

Function TomasQueue.LastElement: Pointer;
  Var
    LastItem: TNode;
  Begin
  LastItem:=FItems.QPred;
  If LastItem = FItems Then
    Begin
    Result:=Nil;
    End
  Else
    Begin
    Result:=LastItem.Item;
    End;
  End;   {TomasQueue.LastElement}

Procedure TomasQueue.Remove(Comp: TomasElement);
  Var
    RemItem: TNode;
    CurWT: Double;
  Begin
  RemItem:=Comp.FQueues.ENext;
  While (RemItem<>Comp.FQueues) And (RemItem.Queue<>Self) Do
    Begin
    RemItem:=RemItem.ENext;
    End;
  If RemItem = Comp.FQueues Then
    Begin
    MessageDlg(TNowString + ' Error(Remove): ' + Comp.Name + ' cannot be removed '
     + ' from ' + Name,mtError,[mbOK],0);
    RunError(255);
    End;
  If ClaimRoom = Self Then
    Comp.FInClaimRoom:=FALSE;
  CalculateMeanLength;
  CurWT:=TNow - RemItem.ArrivalTime;
  FMeanWT:=FTotalPassed * FMeanWT + CurWT;
  If CurWT < FMinWT Then
    Begin
    FMinWT:=CurWT;
    End;
  If CurWT > FMaxWT Then
    Begin
    FMaxWT:=CurWT;
    End;
  Inc(FTotalPassed);
  FMeanWT:=FMeanWT/FTotalPassed;
  If TNow = RemItem.ArrivalTime Then
    Begin
    Inc(FDirectPassed);
    End;
  RemItem.Free;
  Dec(FLength);
  If TomasForm.Trace And DoTrace Then
    CheckTrace(TNowString,Comp.Name,'out of '+ Name,-1);
  If FLength < FMinLength Then
    Begin
    FMinLength:=FLength;
    End;
  End;   {TomasQueue.Remove}

Function TomasQueue.Successor(Comp: TomasElement): Pointer;
  Var
    NextItem: TNode;
  Begin
  NextItem:=Comp.FQueues.ENext;
  While (NextItem<>Comp.FQueues) And (NextItem.Queue<>Self) Do
    Begin
    NextItem:=NextItem.ENext;
    End;
  If NextItem = Comp.FQueues Then
    Begin
    Result:=Nil;
    MessageDlg(TNowString + ' Error(Successor): ' + Comp.Name + ' is not in '
     + Name,mtError,[mbOK],0);
    RunError(255);
    End
  Else
    Begin
    NextItem:=NextItem.QNext;
    If NextItem = FItems Then
      Begin
      Result:=Nil;
      End
    Else
      Begin
      Result:=NextItem.Item;
      End;
    End;
  End;   {TomasQueue.Successor}

Function TomasQueue.Predecessor(Comp: TomasElement): Pointer;
  Var
    PredItem: TNode;
  Begin
  PredItem:=Comp.FQueues.ENext;
  While (PredItem<>Comp.FQueues) And (PredItem.Queue<>Self) Do
    Begin
    PredItem:=PredItem.ENext;
    End;
  If PredItem = Comp.FQueues Then
    Begin
    Result:=Nil;
    MessageDlg(TNowString + ' Error(Predecessor): ' + Comp.Name + ' is not in '
     + Name,mtError,[mbOK],0);
    RunError(255);
    End
  Else
    Begin
    PredItem:=PredItem.QPred;
    If PredItem = FItems Then
      Begin
      Result:=Nil;
      End
    Else
      Begin
      Result:=PredItem.Item;
      End;
    End;
  End;   {TomasQueue.Predecessor}

Function TomasQueue.Contains(Comp: TomasElement): Boolean;
  Var
    ContItem: TNode;
  Begin
  If Comp = Nil Then
    Begin
    Result:=FALSE;
    End
  Else
    Begin
    ContItem:=Comp.FQueues.ENext;
    While (ContItem<>Comp.FQueues) And (ContItem.Queue<>Self) Do
      Begin
      ContItem:=ContItem.ENext;
      End;
    If ContItem = Comp.FQueues Then
      Begin
      Result:=FALSE;
      End
    Else
      Begin
      Result:=TRUE;
      End;
    End;
  End;   {TomasQueue.Contains}

function TomasQueue.Element(Ix: Integer): Pointer;
var
  Ixx: Integer;
  Elm: TomasElement;
begin
  Ixx:=Ix;
  Elm:=FirstElement;
  While (Elm<>Nil) and (Ixx > 0) Do
  Begin
    Elm:=Successor(Elm);
    Dec(Ixx);
  End;
  Result:=Elm;
end;

function TomasQueue.IndexOf(TElm: TomasElement): Integer;
Var
  Ix: Integer;
  Elm: TomasElement;
begin
  Ix:=0;
  Elm:=FirstElement;
  While (Elm<>Nil) and (Elm<>TElm) Do
  Begin
    Inc(Ix);
    Elm:=Successor(Elm);
  End;
  If Elm =Nil Then
    Result:=-1
  Else
    Result:=Ix;
end;

function TomasQueue.ElementWithName(Nm: String): Pointer;
var
  Elm: TomasElement;
begin
  Elm:=FirstElement;
  While (Elm<>Nil) and (Elm.Name <> Nm) Do
  Begin
    Elm:=Successor(Elm);
  End;
  Result:=Elm;
end;

Procedure TomasQueue.CalculateMeanLength;
  Begin
  If TNow <> FResetTime Then
    Begin
    FMeanLength := ((FLastChange - FResetTime) * FMeanLength
     + (TNow - FLastChange) * FLength)/(TNow - FResetTime);
    End;
  FLastChange:=TNow;
  End;   {TomasQueue.CalculateMeanLength}

Procedure TomasQueue.ResetStatistics;
  Begin
  FMeanLength:=FLength;
  FMinLength:=FLength;
  FMaxLength:=FLength;
  FMeanWT:=0;
  FMinWT:=NEVER;
  FMaxWT:=0;
  FDirectPassed:=0;
  FTotalPassed:=0;
  FResetTime:=TNow;
  End;  {TomasQueue.ResetStatistics}

Function TomasQueue.Copy: TomasQueue;
  Var
    NewQueue: TomasQueue;
    Elm: TomasElement;
  Begin
  NewQueue:=TomasQueue.Create('Copy_' + FName);
  Elm:=FirstElement;
  While Elm<>Nil Do
    Begin
    Elm.EnterQueue(NewQueue);
    Elm:=Successor(Elm);
    End;
  Result:=NewQueue;
  End;   {TomasQueue.Copy}

Procedure TomasQueue.Clear;
  Var
    Elm: TomasElement;
  Begin
  Elm:=FirstElement;
  While Elm<>Nil Do
    Begin
    Remove(Elm);
    Elm:=FirstElement;
    End;
  End;   {TomasQueue.Clear}

Function TomasQueue.Move: TomasQueue;
  Var
    NewQueue: TomasQueue;
  Begin
  NewQueue:=Copy;
  Clear;
  Result:=NewQueue;
  End;  {TomasQueue.Move}


Constructor TMsgString.Create(Msg: String);
  Begin
  inherited Create;
  MsgS:=Trim(Msg);
  SubMsgS:='';
  End;  {TMsgString.Create}

Procedure TMsgString.SetNextStringPart;
  Var
    P: Integer;
  Begin
  P:=Pos('!!',MsgS);
  If P = 0 Then
    Begin
    SubMsgS:=Trim(MsgS);
    MsgS:='';
    End
  Else
    Begin
    SubMsgS:=Trim(Copy(MsgS,1,P-1));
    MsgS:=Copy(MsgS,P+2,Length(MsgS) - P - 1);
    End;
  MsgError:=FALSE;
  End;

Function TMsgString.GetBoolean: Boolean;   {TRUE if y,Y,t,T,1}
  Var
    C: Char;
  Begin
  SetNextStringPart;
  C:=SubMsgS[1];
  Case C Of
    'y','Y','t','T','1':
      Begin
      Result:=TRUE;
      End;
    Else
      Begin
      Result:=FALSE;
      End;
    End;
  End;  {TMsgString.GetBoolean}

Function TMsgString.GetSmallInt: SmallInt; {-32768,..,32767}
  var
    II: Integer;
  Begin
  II:=GetInteger;
  If (II>32767) Or (II<-32768) Then
    Begin
    MsgError:=TRUE;
    II:=0;
    End;
  Result:=II;
  End;  {TMsgString.GetSmallInt}

Function TMsgString.GetLongInt: LongInt;   {-2147483648,..,2147483647}
  Begin
  Result:=GetInteger;
  End;  {TMsgString.GetLongInt}

Function TMsgString.GetByte: Byte;         {0,..,255}
  Var
    II: Integer;
  Begin
  II:=GetInteger;
  If (II > 255) Or (II < 0) Then
    Begin
    MsgError:=TRUE;
    II:=0;
    End;
  Result:=II;
  End;  {TMsgString.GetByte}

Function TMsgString.GetWord: Word;         {0,..,65535}
  Var
    II: Integer;
  Begin
  II:=GetInteger;
  If (II > 65535) Or (II < 0) Then
    begin
    MsgError:=TRUE;
    II:=0;
    End;
  Result:=II;
  End;  {TMsgString.GetWord}

Function TMsgString.GetInteger: Integer;   {-2147483648,..,2147483647}
  Var
    I: Integer;
  Begin
  SetNextStringPart;
  If SubMsgS = '' Then
    begin
    Result:=0;
    MsgError:=TRUE;
    End
  Else
    Begin
    Try
      I:=StrToInt(SubMsgS);
    Except
      On E: EConvertError Do
        Begin
        MsgError:=TRUE;
        I:=0;
        End;
      End;
    Result:=I;
    End;
  End;  {TMsgString.GetInteger}

Function TMsgString.GetCardinal: Cardinal; {0,..,2147483647}
  Var
    II: Integer;
  Begin
  II:=GetInteger;
  If II < 0 Then
    Begin
    MsgError:=TRUE;
    II:=0;
    End;
  Result:=II;
  End;  {TMsgString.GetCardinal}

Function TMsgString.GetReal: Real;         {6 bytes}
  Var
    D: Double;
  Begin
  D:=GetDouble;
  If (D > 1.7E+38) Or (D < -1.7E+38) Then
    Begin
    MsgError:=TRUE;
    Result:=0;
    End
  Else
    begin
    Result:=D;
    End;
  End;  {TMsgString.GetReal}

Function TMsgString.GetDouble: Double;     {8 bytes}
  Var
    D: Double;
  Begin
  SetNextStringPart;
  If SubMsgS = '' Then
    begin
    MsgError:=TRUE;
    D:=0;
    End
  Else
    Begin
    Try
      D:=StrToFloat(SubMsgS);
    Except
      D:=0;
      MsgError:=TRUE;
      End;
    End;
  Result:=D;
  End;  {TMsgString.GetDouble}

Function TMsgString.GetSingle: Single;     {4 bytes}
  Var
    D: Single;
  Begin
  SetNextStringPart;
  If SubMsgS = '' Then
    begin
    MsgError:=TRUE;
    D:=0;
    End
  Else
    Begin
    Try
      D:=StrToFloat(SubMsgS);
    Except
      D:=0;
      MsgError:=TRUE;
      End;
    End;
  Result:=D;
  End;  {TMsgString.GetSingle}

Function TMsgString.GetString: String;
  Begin
  SetNextStringPart;
  Result:=SubMsgS;
  End;  {TMsgString.GetString}


(**************************************************************)
(*                   TNode methods                            *)
(**************************************************************)

Destructor TNode.Destroy;
  Begin
  If (EPred<>Nil) Then
    Begin
    {remove ref from Elementchain}
    EPred.ENext:=ENext;
    ENext.EPred:=EPred;
    End;
  If (QPred <> Nil) Then
    Begin
    {remove ref from Queuechain}
    QPred.QNext:=QNext;
    QNext.QPred:=QPred;
    End;
  inherited Destroy;
  End;

Procedure TNode.ElementToQueueBefore(Tail: TNode);
  Begin
  QPred:=Tail.QPred;
  QNext:=Tail;
  QPred.QNext:=Self;
  Tail.QPred:=Self;
  End;

Procedure TNode.ElementToQueueAfter(Head: TNode);
  begin
  QNext:=Head.QNext;
  QPred:=Head;
  QNext.QPred:=Self;
  Head.QNext:=Self;
  End;

Procedure TNode.QueueToElementBefore(Tail: TNode);
  Begin
  EPred:=Tail.EPred;
  ENext:=Tail;
  EPred.ENext:=Self;
  Tail.EPred:=Self;
  End;


(**************************************************************)
(*                   TQueueLink methods                       *)
(**************************************************************)

Destructor TQueueLink.Destroy;
  Begin
  Pred.Next:=Next;
  Next.Pred:=Pred;
  inherited destroy;
  End;


(**************************************************************)
(*                    Distribution methods                    *)
(**************************************************************)

Const
  Multiplier: Int64  = 16807;

Function Rand(Var L: Integer): Double;
  Begin
  L:=((Int64(L) * Multiplier) AND $7fffffff) mod $80000000;
  Result:=L/$7fffffff;
  End;  {Rand}

Constructor TDistribution.Create(DSeed: Integer);
  Begin
  Inherited Create;
  Seed:=DSeed;
  DMean:=0;
  End;      {TDistribution.Create}

Function TDistribution.Mean: Double;
   Begin
   Result := DMean;
   End;     {TDistribution.Mean}

Function TDistribution.Sample: Double;
  Begin
  Result:=Rand(Seed);
  End;   {TDistribution.Sample}

Function TDistribution.CurrentSeed: Integer;
  Begin
  Result:=Seed;
  End;   

Constructor TExponentialDistribution.Create(Seed: Integer; Mn: Double);
  Begin
  inherited Create(Seed);
  DMean:=Mn;
  End;      {TExponentialDistribution.Init}

Function TExponentialDistribution.Sample: Double;
  Begin
  Result:=-DMean * Ln(Rand(Seed));
  End;   {TExponentialDistribution.Sample}

Constructor TUniformDistribution.Create(Seed: Integer; Lb,Ub: Double);
  Begin
  inherited Create(Seed);
  LBound:=Lb;
  UBound:=Ub;
  DMean:=(Lb+Ub)/2.0;
  End;      {TUniformDistribution.Init}

Function TUniformDistribution.Sample: Double;
  Begin
  Result:=LBound+Rand(Seed)*(UBound-LBound);
  End;   {TUniformDistribution.Sample}

Constructor TNormalDistribution.Create(Seed: Integer; Mn,S: Double);
  Begin
  inherited Create(Seed);
  DMean:=Mn;
  Sigma:=S;
  End; {TNormalDistribution.Create}

Function TNormalDistribution.Sample: Double;
  Var
    U,V,R: Double;
    Accepted: Boolean;
  Begin
  Accepted:=FALSE;
  R:=1;
  While NOT(Accepted) Do
    Begin
    U := inherited Sample;
    V := inherited Sample;
    V := 1.7156 * (V - 0.5);
    R := V/U;
    Accepted := (R*R <= -4.0*Ln(U));
    End;
  Result:=Sigma * R + DMean;
  End;  {TNormalDistribution.Sample}

Constructor TTableDistribution.Create(Seed: Integer; DType: DistributionTableType);
  Begin
  Inherited Create(Seed);
  TableType:=DType;
  XFValues:=TList.Create;
  DMean := 0;
  End;   { TTableDistribution.Create}

Function XOnlySort(Item1,Item2: Pointer): Integer;
 Begin
 If TXFItem(Item1).X_Value < TXFItem(Item2).X_Value Then
   Begin
   Result:=-1;
   End
 Else
   begin
   If TXFItem(Item1).X_Value > TXFItem(Item2).X_Value Then
     Begin
     Result:=1;
     End
   Else
     Begin
     Result:=0;
     End;
   End;
 End; {XOnlySort}

Function CumulativeSort(Item1,Item2: Pointer): Integer;
  Begin
  If TXFItem(Item1).X_Value < TXFItem(Item2).X_Value Then
    Begin
    Result:=-1;
    End
  Else
    begin
    If TXFItem(Item1).X_Value > TXFItem(Item2).X_Value Then
      Begin
      Result:=1;
      End
    Else
      Begin
      If TXFItem(Item1).F_Value < TXFItem(Item2).F_Value Then
        Begin
        Result:=-1;
        End
      Else
        Begin
        If TXFItem(Item1).F_Value > TXFItem(Item2).F_Value Then
          Begin
          Result:=1;
          End
        Else
          Begin
          Result:=0;
          End;
        End;
      End;
    End;
  End;     {CumulativeSort}

Procedure TTableDistribution.AddValue(X,F: Double);
  Var
    X0, X1, XFValue1: TXFItem;
    FSum: Double;
    i: Integer;
  Begin
  If (F < 0.0) Or (F > 1.0) then
    Begin
    MessageDlg('Error(AddValue): Cumulative value or probability out of range [0,1]',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  If TableType = Discrete Then
    Begin
    FSum:=0;
    For i:=0 To XFValues.Count - 1 Do
      Begin
      XFvalue1:=XFValues.Items[i];
      FSum:=FSum + XFValue1.F_Value;
      End;
    If Round(100 * (FSum + F)) /100 > 1 Then
      Begin
      MessageDlg('Error(AddValue): Sum of probabilities becomes > 1.0',
       mtError,[mbOK],0);
      Runerror(255);
      End;
    end;
  XFValue1:=TXFItem.Create;
  XFValue1.X_Value:=X;
  XFValue1.F_Value:=F;
  XFValues.Add(XFValue1);
  If XFValues.Count > 1 Then
    Begin
    If TableType = Cumulative Then
      begin
      XFValues.Sort(CumulativeSort);
      End
    Else
      Begin
      XFValues.Sort(XOnlySort);
      End;
    End;
   DMean:= 0;
   FSum:= 0;
   For i := 1 To XFValues.Count - 1 Do
     Begin
     X0 := XFValues.Items[i-1];
     X1 := XFValues.Items[i];
     If TableType = Discrete Then
       Begin
       DMean:= DMean + X1.F_Value * (X0.X_Value + X1.X_Value) / 2;
       FSum:= FSum + X1.F_Value;
       End
     Else
       Begin
       DMean:= DMean + (X1.F_Value-X0.F_Value) * (X0.X_Value + X1.X_Value) / 2;
       FSum:= X1.F_Value;
       End;
     End;
   If FSum > 0 Then
     Begin
     DMean := DMean/FSum;
     End;
  End;   {TTableDistribution.AddValue}

Function TTableDistribution.Sample: Double;
  Var
    F: Double;
    FSum: Double;
    XFV1: TXFItem;
    XFV2: TXFItem;
    I: Integer;
  Begin
  If XFValues.Count = 0 Then
    Begin
    MessageDlg('Error(Sample): Try to sample from an empty table',
     mtError,[mbOK],0);
    Runerror(255);
    End;
  F:=Rand(Seed);
  If TableType = Cumulative Then
    Begin
    I:=0;
    While (I <= XFValues.Count - 1) And (F > TXFItem(XFValues.Items[I]).F_Value) Do
      Begin
      Inc(I);
      End;
    If I = 0 Then
      Begin
      Result:=TXFItem(XFValues.Items[0]).X_Value;
      End
    Else
      Begin
      If I = XFValues.Count Then
        Begin
        Result:=TXFItem(XFValues.Items[I-1]).X_Value;
        End
      Else
        Begin
        XFV1:=XFValues.Items[I-1];
        XFV2:=XFValues.Items[I];
        F:=(F - XFV1.F_Value)/(XFV2.F_Value - XFV1.F_Value);
        Result:=XFV1.X_Value + F * (XFV2.X_Value - XFV1.X_Value);
        End;
      End;
    End
  Else
    Begin
    FSum:=TXFItem(XFValues.Items[0]).F_Value;
    I:=0;
    While (I < XFValues.Count - 1) And (F > FSum) Do
      Begin
      Inc(I);
      FSum:=FSum + TXFItem(XFValues.Items[I]).F_Value;
      End;
    Result:=TXFItem(XFValues.Items[I]).X_Value;
    End;
  End;   {TTableDistribution.Sample}

Constructor TInputDistribution.Create(DString: String);
  { InputString must have one of the formats:
    'Uniform(Parm1,Parm2),Seed'
    'Exponential(Parm1),Seed'
    'Normal(Parm1,Parm2),Seed'
    'Table(DistributionTableType,(X1,F1,X2,F2,...,Xn,Fn)),Seed'.
    Only the first 3 characters of the distributiontype are significant
  }
    Var
      i: Integer;
      j: Integer;
      S: String;
      X: Double;
      F: Double;
    Const
      Dist: Array[1..4] Of String[3] = ('UNI','EXP','NOR','TAB');

    Function GetSeed: Integer;
      var
        P: Integer;
      Begin      {procedure works only for correct strings}
      P:=Pos('),',DString);
      Result:=StrToInt(Copy(Dstring,P+2,Length(DString) - P - 1));
      End;

    Function GetParm(I: Integer): String;
      Var
        P: Integer;
        PP: Integer;
        S: String;
      Begin            {procedure works only for correct strings}
      P:=Pos('(',DString);
      S:=Copy(Dstring,P+1,Length(DString) - P);  {remove type string}
      P:=Pos(',',S);
      If I =1 Then
        begin
        If S[P-1] = ')' Then
          Begin      {Only 1 parameter}
          Result:=Copy(S,1,P-2);
          End
        Else
          Begin
          Result:=Copy(S,1,P-1);
          End;
        End
      Else
        Begin
        PP:=Pos(')',S);
        Result:=Copy(S,P+1,PP-P-1);
        End;
      End;

    Begin
    DString:=UpperCase(DString);
    i:=5;
    j:=0;
    While (i > 1) And (j = 0) Do
      begin
      Dec(i);
      j:=Pos(Dist[i],DString);
      End;
    Case i of
      1:Begin
        InputType:=Uniform;
        Distribution:=TUniformDistribution.Create(GetSeed,
         StrToFloat(GetParm(1)),StrToFloat(GetParm(2)));
        End;
      2:Begin
        InputType:=Exponential;
        Distribution:=TExponentialDistribution.Create(GetSeed,
         StrToFloat(GetParm(1)));
        End;
      3:Begin
        InputType:=Normal;
        Distribution:=TNormalDistribution.Create(GetSeed,
         StrToFloat(GetParm(1)),StrToFloat(GetParm(2)));
        End;
      4:Begin
        InputType:=Table;
        S:=GetParm(1);
        If UpperCase(S) = 'DISCRETE' Then
          Begin
          Distribution:=TTableDistribution.Create(GetSeed,Discrete);
          End
        Else
          Begin
          Distribution:=TTableDistribution.Create(GetSeed,Cumulative);
          End;
        i:=Pos(',(',DString);
        j:=Pos('))',DString);
        S:=Copy(DString,i+2,j - i -2);
        While S <> '' Do
          Begin
          i:=Pos(',',S);
          X:=StrToFloat(Copy(S,1,i-1));
          S:=Copy(S,i+1,Length(S) - i);
          i:=Pos(',',S);
          If i > 0 Then
            Begin
            F:=StrToFloat(Copy(S,1,i-1));
            S:=Copy(S,i+1,Length(S) - i);
            End
          Else
            Begin
            F:=StrToFloat(S);
            S := '';
            End;
          TTableDistribution(Distribution).AddValue(X,F);
          End;
        End;
      Else
        Begin
        MessageDlg('Error(InputDistribution): Unknown DistributionType',
         mtError,[mbOK],0);
        RunError(255);
        End;
      End;
    End; {TInputDistribution.Create}

Function TInputDistribution.Mean: Double;
   Begin
   Result := Distribution.Mean;
   End;    {TInputDistribution.Mean}

  Function TInputDistribution.Sample: Double;
    Begin
    Result:=Distribution.Sample;
    End;  {TInputDistribution.Sample}

Procedure StatusChange;
  var
    NewQuestion: TMsgString;
    CurrentStatus: SimulationStatus;
    Phrase: String;
    Source: String;
    S: String;
  Begin
  if ServerQuestion And Assigned(PAnswerQuestion) Then
    Begin
    QuestionParms(Phrase,Source,QuestionID);
    NewQuestion:=TMsgString.Create(Phrase);
    PAnswerQuestion(NewQuestion,Source);
    NewQuestion.Free;
    End;
  If ServerMessage And Assigned(PReceiveMessage) Then
    Begin
    PReceiveMessage(MessageLine);
    End;
  If TomasForm.Trace And TracingMessage Then
    Begin
    WriteToTrace(TraceLine);
    End;
  If CurrentChange And TomasForm.Trace Then
    Begin
    If TomasForm.ClockTimeBox.Checked Then
      S:=TimeToStr(Time) + '/'
    Else
      S:='';
    WriteToTrace('');  
    WriteToTrace(S + TNowString + '  ' +
     CurrentElement.Name + ' is current now');
     End;
  If TNow - LastEditTime >= TimeDisplayInterval Then
    Begin
    if TomasForm.Visible Then
     TomasForm.TomasStatusBar.Panels[0].Text:='Time: ' + Trim(TNowString);
    If Assigned(PTimer) Then
      Begin
      PTimer;
      End;
    LastEditTime:=TNow;
    End;
  CurrentStatus:=GetSimulationStatus;
  Case CurrentStatus Of
    S_INIT:
      Begin
      End;
    S_RUNNING:
      Begin
      If Assigned(TomasAnimationForm) And (TomasAnimationForm.Visible) Then
        Begin
        If TNow - LastRefreshTime > AnimationDisplayInterval Then
          Begin
          TomasAnimationForm.SetFocus;
          TScreenString:=Trim(TNowString) + ' ' + ScreenString;
          TomasAnimationForm.Refresh(Nil);
          LastRefreshTime:=TNow;
          End;
        End;
      End;
    S_STEPRUNNING:
      Begin
      End;
    S_INTERRUPTED:
      Begin
      WriteToTrace(TNowString + ' *** Interrupt ***');
      TomasForm.ResumeItem.Enabled:=TRUE;
      TomasForm.InterruptItem.Enabled:=FALSE;
      If GetBreakTime = -1 Then
        Begin
        TomasForm.HeaderControl1.Sections[1].Text:='Break: ';
        End;
      End;
    S_FINISHED:
      Begin
      End;
    End;
  End;

Procedure ResetTomas;
  Var
    CurClass: TElementClass;
    QLink: TQueueLink;
    Q: TomasQueue;
    CurElement: TomasElement;
  Begin
  If RunStatus = RunFinished Then
    Begin
    TomasForm.SetTrace(FALSE);
    TomasForm.SetStepMode(FALSE);
    TomasForm.StepButton.Hide;
    CurClass:=ElementClassHead;
    While CurClass<>Nil Do
      Begin
      ElementClassHead:=CurClass.Next;
      CurClass.Free;
      CurClass:=ElementClassHead;
      End;
    ElementClassTail:=Nil;
    DoTrace:=FALSE;
    QLink:=QueueRoom.Next;
    While QLink <> QueueRoom Do
      Begin
      Q:=QLink.Ref;
      Q.Free;
      QLink:=QueueRoom.Next;
      End;
    CurElement:=ElementRoom.FirstElement;
    While CurElement<>Nil Do
      Begin
      CurElement.Free;
      CurElement:=ElementRoom.FirstElement;
      End;
    NrOfQueues:=0;
    LastEditTime:=0;
    TimeDisplayInterval:=1;
    ResetTomasProcess;
    End;
  End;


(**************************************************************)
(*                    Initialization                          *)
(**************************************************************)

Procedure InitializeTomas;
  Begin
  Application.CreateForm(TTomasForm,TomasForm);
  SetStatusChange(StatusChange);
  TomasForm.SetTrace(FALSE);
  TomasForm.SetStepMode(FALSE);
  TomasForm.StepButton.Hide;
  SetModelName('');
  ElementClassHead:=Nil;
  ElementClassTail:=Nil;
  QueueRoom:=TQueueLink.Create;
  QueueRoom.Ref:=Nil;
  QueueRoom.Next:=QueueRoom;
  QueueRoom.Pred:=QueueRoom;
  NrOFQueues:=0;
  DoTrace:=FALSE;
  ElementRoom:=TomasQueue.Create('Elements');
  ElementRoom.FQLink.Free;  {we don't need it in the QueueRoom}
  ClaimRoom:=TomasQueue.Create('Claimers');
  ClaimRoom.FQLink.Free;    {we don't need it in the Queueroom}
  DoTrace:=TRUE;
  OutputFilename:='';
  OutputDestinationFile:=FALSE;
  LastEditTime:=0;
  TimeDisplayInterval:=1;
  End;

(**************************************************************)
(*                    TomasForm methods                       *)
(**************************************************************)

procedure TTomasForm.GoCreateTomasForm(Sender: TObject);
begin
StartItem.Enabled:=TRUE;
InterruptItem.Enabled:=FALSE;
ResumeItem.Enabled:=FALSE;
AnimationOn:=FALSE;
HeaderControl1.Sections[0].Text:='Server Address:';
HeaderControl1.Sections[1].Text:='Break:';
HeaderControl1.Sections[2].Text:='Output: Screen';
HeaderControl1.Sections[3].Text:='Animation: Off';
HeaderControl1.Sections[4].Text:='Message Log: Off';
QuitItem.Enabled:=TRUE;
TracePanel.Width:=ClientWidth div 2;
TracePanel.Height:=StepBevel.Height + 10;;
TracePanel.Top:=HeaderControl1.Top + HeaderControl1.Height;
TracePanel.Left:=0;
Tracememo.Hide;
TraceCheckbox.Checked:=FALSE;
StepModeCheckBox.Checked:=FALSE;
CloseWindow:=FALSE;
LastRefreshTime:=-1;
end;

Procedure TTomasForm.SetTraceMemo;
  Begin
  if FTrace or FTraceStandBy Then
    Begin
    TracePanel.Height:=ClientHeight - TomasStatusBar.Height;
    TraceMemo.Width:=TracePanel.Width - 10;
    TraceMemo.Left:=5;
    TraceMemo.Top:=StandbyCheckBox.Top + 30;
    TraceMemo.Height:=TracePanel.Height - TraceMemo.Top - 5;
    Tracememo.Show;
    End
  Else
    Begin
    TracePanel.Height:=StepBevel.Height + 10;
    TraceMemo.Hide;
    End;
  End;

Procedure TTomasForm.SetTrace(OnOff: Boolean);
  Begin
  FTrace:=OnOff;
  TraceCheckBox.Checked:=OnOff;
  SetTraceMemo;
  End;

Procedure TTomasForm.SetStepMode(OnOff: Boolean);
  Begin
  FStepMode:=OnOff;
  StepModeCheckBox.Checked:=OnOff;
  StepButton.Visible:=OnOff;
  if OnOff Then
    Begin
    StartStepMode;
    End
  Else
    Begin
    StopStepMode;
    End;
  End;

Function TTomasForm.GetServerAddress: String;
  Var
    P: Integer;
  Begin
  P:=Pos(':',HeaderControl1.Sections[0].Text);
  Result:=Trim(Copy(HeaderControl1.Sections[0].Text,P+1,
   Length(HeaderControl1.Sections[0].Text) - P));
  End;

Procedure TTomasForm.SetServerAddress(SrvAddress: String);
  Begin
  HeaderControl1.Sections[0].Text:='ServerAddress: ' + SrvAddress;
  SetServer(SrvAddress);
  End;

procedure TTomasForm.TraceCheckClick(Sender: TObject);
begin
  SetTrace(TraceCheckBox.Checked);
end;

procedure TTomasForm.StepModeCheckClick(Sender: TObject);
begin
SetStepMode(StepModeCheckBox.Checked);
end;

procedure TTomasForm.ExecuteNextStep(Sender: TObject);
begin
PerformNextEvent;
end;

procedure TTomasForm.GoHideSettings(Sender: TObject);
begin
SettingsPanel.Hide;
end;

Procedure TTomasForm.GoSetAnimationOnOrOff;
  Begin
  If AnimationOn Then
    Begin
    HeaderControl1.Sections[3].Text:='Animation: On';
    If NOT(Assigned(TomasAnimationForm)) Then
      Begin
      Application.CreateForm(TTomasAnimationForm,TomasAnimationForm);
      Width:=120;
      Top:=0;
      Left:=0;
      TomasAnimationForm.Left:=Width;
      TomasAnimationForm.Top:=0;
      TomasAnimationForm.Width:=Screen.Width - Width;
      TomasAnimationForm.Height:=Screen.Height;
      End;
    TomasAnimationForm.Show;
    TomasAnimationForm.BringToFront;
    End
  Else
    Begin
    HeaderControl1.Sections[3].Text:='Animation: Off';
    If Assigned(TomasAnimationForm) Then
      Begin
      TomasAnimationForm.Hide;
      End;
    End;
  End;

procedure TTomasForm.GoAcceptSettings(Sender: TObject);
Var
  S: String;
  T: Double;
begin
If ServerAddressCombo.Enabled Then
  Begin
  S:=Trim(ServerAddressCombo.Text);
  If (Uppercase(S) = 'NONE') Or (S='') Then
    Begin
    SetServer('');
    HeaderControl1.Sections[0].Text:='ServerAddress: ';
    End
  Else
    Begin
    SetServer(S);
    HeaderControl1.Sections[0].Text:='ServerAddress: ' + S;
    End;
  End;
S:=Trim(BreakEdit.Text);
If S='' Then
  begin
  T:=-1;
  End
Else
  Begin
  T:=StrToFloat(S);
  End;
SetBreakTime(T);
If T>=TNow Then
  Begin
  HeaderControl1.Sections[1].Text:='Break: ' + S;
  End
Else
  Begin
  HeaderControl1.Sections[1].Text:='Break: ';
  End;
S:=Trim(OutputCombo.Text);
If (Uppercase(S) = 'SCREEN') Or (S = '') Then
  Begin
  OutputDestinationFile:=FALSE;
  OutputFileName:='Screen';
  End
Else
  Begin
  AssignFile(OutputFile,S);
  {$I-}
  If NOT(FileExists(S)) Then
    Begin
    ReWrite(OutputFile);
    End
  Else
    Begin
    Reset(OutputFile);
    End;
  {$I+}
  If IOResult <> 0 Then
    Begin
    MessageDlg('Error opening file ' + S,mtWarning,[mbOK],0);
    OutputDestinationFile:=FALSE;
    OutputFileName:='Screen';
    End
  Else
    Begin
    OutputDestinationFile:=TRUE;
    CloseFile(OutputFile);
    OutputFileName:=S;
    End;
  End;
HeaderControl1.Sections[2].Text:='Output: ' + OutputFileName;
AnimationOn:=AnimationCheckBox.Checked;
GoSetAnimationOnOrOff;
If MessageCheckBox.Checked Then
  Begin
  DoStartLogging;
  HeaderControl1.Sections[4].Text:='Message Log: On';
  End
Else
  Begin
  DoStopLogging;
  HeaderControl1.Sections[4].Text:='Message Log: Off';
  End;
SettingsPanel.Hide;
end;

Procedure SetOutputDestination(S: String);
Begin
S:=Trim(S);
If (Uppercase(S) = 'SCREEN') Or (S = '') Then
  Begin
  OutputDestinationFile:=FALSE;
  OutputFileName:='Screen';
  End
Else
  Begin
  AssignFile(OutputFile,S);
  {$I-}
  If NOT(FileExists(S)) Then
    Begin
    ReWrite(OutputFile);
    End
  Else
    Begin
    Reset(OutputFile);
    End;
  {$I+}
  If IOResult <> 0 Then
    Begin
    MessageDlg('Error opening file ' + S,mtWarning,[mbOK],0);
    OutputDestinationFile:=FALSE;
    OutputFileName:='Screen';
    End
  Else
    Begin
    OutputDestinationFile:=TRUE;
    CloseFile(OutputFile);
    OutputFileName:=S;
    End;
  End;
End;

procedure TTomasForm.CheckFormResize(Sender: TObject);
begin
TracePanel.Width:=ClientWidth div 2;
TracePanel.Top:=HeaderControl1.Top + HeaderControl1.Height;
TracePanel.Left:=0;
SetTraceMemo;
If Assigned(ElementRoom) Then
  Begin
  TomasStatusBar.Panels[1].Text:='Elements: ' + IntToStr(ElementRoom.Length);
  TomasStatusBar.PAnels[2].Text:='Queues: ' + IntToStr(NrOfQueues);
  TomasStatusBar.Panels[0].Text:='Time: ' + Trim(TNowString);
  End;
end;

procedure TTomasForm.GoStartSimulation(Sender: TObject);
begin
TomasForm.InterruptItem.Enabled:=TRUE;
TomasForm.StartItem.Enabled:=FALSE;
StartSimulation;
If CloseWindow Then
  Begin
  Close;
  End;
end;

procedure TTomasForm.GoResumeSimulation(Sender: TObject);
begin
TomasForm.InterruptItem.Enabled:=TRUE;
TomasForm.ResumeItem.Enabled:=FALSE;
ResumeProcessSimulation;
end;

procedure TTomasForm.GoQuitSimulation(Sender: TObject);
begin
SetStopTime(TNow);
end;

procedure TTomasForm.CheckClosing(Sender: TObject;
  var Action: TCloseAction);
begin
SetStopTime(TNow);
If CloseWindow Or (GetSimulationStatus <> S_RUNNING) Then
  Begin
  Action:=caFree;
  If ConnectedToServer Then
    DisconnectFromServer;
  If Assigned(TheGraphicsForm) Then
    Begin
    TheGraphicsForm.Close;
    End;
  If GetSimulationStatus <> S_RUNNING Then
    Begin
    ResumeProcessSimulation;
    End;
  End
Else
  Begin
  CloseWindow:=TRUE;
  Action:=caNone;
  StopStepMode;  {be sure it will be not stepping}
  If GetSimulationStatus <> S_RUNNING Then
    Begin
    ResumeProcessSimulation;
    End;
  End;
end;

procedure TTomasForm.GoInterruptSimulation(Sender: TObject);
begin
If GetSimulationStatus > S_INTERRUPTED Then
  Begin
  ResumeItem.Enabled:=TRUE;
  InterruptItem.Enabled:=FALSE;
  InterruptProcessSimulation;
  End;
end;

procedure TTomasForm.CheckSpaceKey(Sender: TObject; var Key: Char);
begin
If Key = ' ' Then
  Begin
  SetKeyPressed(' ');
  End;
end;

procedure TTomasForm.CheckStatusBarMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  Function PanelLeft(I: Integer): Integer;
    Var
      L: Integer;
    Begin
    L:=0;
    While I > 0 Do
      Begin
      L:=L + TomasStatusBar.Panels[I-1].Width;
      Dec(I);
      End;
    Result:=L;
    End;

begin
If Button = mbLeft Then
  Begin
  If X > PanelLeft(1) Then
    Begin
    If X < PanelLeft(2) Then
      Begin
      GoShowElements;
      End
    Else
      Begin
      If X < PanelLeft(3) Then
        Begin
        GoShowQueues;
        End;
      End;
    End;
  End;
end;



procedure TTomasForm.GoResizeTraceMemo(Sender: TObject);
begin
TraceMemo.Width:=TracePanel.Width - 10;
TraceMemo.Left:=5;
TraceMemo.Top:=TraceCheckBox.Top + 30;
TraceMemo.Height:=TracePanel.Height - TraceMemo.Top - 5;
end;

procedure TTomasForm.GoEditTheSettings(HeaderControl: THeaderControl;
  Section: THeaderSection);
Var
  S: String;
  T: Double;
  St: SimulationStatus;
begin
S:=GetServer;
If S='' Then
  begin
  ServerAddressCombo.Text:='None';
  End
Else
  Begin
  ServerAddressCombo.Text:=S;
  End;
St:=GetSimulationStatus;
ServerAddressCombo.Enabled:=(St=S_INIT);{(GetSimulationStatus = S_INIT);}
T:=GetBreakTime;
If T = -1 Then
  begin
  BreakEdit.Text:='';
  End
Else
  Begin
  BreakEdit.Text:=FormatFloat('0.00',T);
  End;
AnimationCheckBox.Checked:=AnimationOn;
MessageCheckBox.Checked:=GetLogging;
MessageCheckBox.Enabled:=(GetServer<>'');
OutputCombo.Text:=OutputFilename;
SettingsPanel.Show;
SettingsPanel.BringTofront;
GoInterruptSimulation(Nil);
end;

procedure TTomasForm.GoSetMessageCheck(Sender: TObject);
begin
If (Trim(ServerAddressCombo.Text) = '') Or (Trim(UpperCase(ServerAddressCombo.Text)) = 'NONE')
 Then
  Begin
  MessageCheckBox.Enabled:=FALSE;
  End
Else
  Begin
  MessageCheckBox.Enabled:=TRUE;
  End;
end;

procedure TTomasForm.GoShowTomasForm(Sender: TObject);
begin
If Assigned(ElementRoom) Then
  Begin
  TomasStatusBar.Panels[1].Text:='Elements: ' + IntToStr(ElementRoom.Length);
  TomasStatusBar.PAnels[2].Text:='Queues: ' + IntToStr(NrOfQueues);
  TomasStatusBar.Panels[0].Text:='Time: ' + Trim(TNowString);
  End;
end;

procedure TTomasForm.TraceStandByCheckClick(Sender: TObject);
begin
  SetTraceStandBy(StandByCheckBox.Checked);
end;

Procedure TTomasForm.SetTraceStandBy(OnOff: Boolean);
begin
  FTraceStandBy:=OnOff;
  StandByCheckBox.Checked:=OnOff;
  SetTraceMemo;
end;

Initialization
  InitializeTomas;

end.
